from __future__ import annotations

from typing import Any, Dict, Sequence, Type

import logging.handlers
import time
import traceback

import dask.distributed
import numpy as np
from sklearn.utils.validation import check_random_state
from smac.callbacks import IncorporateRunResultCallback
from smac.optimizer.smbo import SMBO
from smac.runhistory.runhistory import RunInfo, RunValue
from smac.tae.base import StatusType

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.ensemble_building.builder import EnsembleBuilder
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.ensembles.ensemble_selection import EnsembleSelection
from autosklearn.metrics import Scorer
from autosklearn.util.logging_ import get_named_client_logger


class EnsembleBuilderManager(IncorporateRunResultCallback):
    def __init__(
        self,
        backend: Backend,
        dataset_name: str,
        task: int,
        metrics: Sequence[Scorer],
        time_left_for_ensembles: float = np.inf,
        max_iterations: int | None = None,
        pynisher_context: str = "fork",
        ensemble_class: Type[AbstractEnsemble] = EnsembleSelection,
        ensemble_kwargs: Dict[str, Any] | None = None,
        ensemble_nbest: int | float = 50,
        max_models_on_disc: int | float | None = None,
        seed: int = 1,
        precision: int = 32,
        memory_limit: int | None = None,
        read_at_most: int | None = None,
        logger_port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
        random_state: int | np.random.RandomState | None = None,
        start_time: float | None = None,
    ):
        """SMAC callback to handle ensemble building

        Parameters
        ----------
        backend: Backend
            backend to write and read files

        dataset_name: str
            name of dataset

        task: int
            Type of ML task

        metrics: Sequence[Scorer]
            Metrics to optimize the ensemble for

        time_left_for_ensemble: float = np.inf
            How much time is left for the task in seconds.
            Job should finish within this allocated time

        max_iterations: int | None = None
            maximal number of iterations to run this script. None indicates no limit
            on iterations.

        pynisher_context: "spawn" | "fork" | "forkserver" = "fork"
            The multiprocessing context for pynisher.

        ensemble_class : Type[AbstractEnsemble] (default=EnsembleSelection)
            Class implementing the post-hoc ensemble algorithm. Set to
            ``None`` to disable ensemble building or use ``SingleBest``
            to obtain only use the single best model instead of an
            ensemble.

        ensemble_kwargs : Dict, optional
            Keyword arguments that are passed to the ensemble class upon
            initialization.

        ensemble_nbest: int | float = 50
            If int: consider only the n best prediction
            If float: consider only this fraction of the best models

        max_models_on_disc: int | float | None = None
           Defines the maximum number of models that are kept in the disc.

           If int, it must be greater or equal than 1, and dictates the max
           number of models to keep.

           If float, it will be interpreted as the max megabytes allowed of
           disc space. That is, if the number of ensemble candidates require more
           disc space than this float value, the worst models will be deleted to
           keep within this budget. Models and predictions of the worst-performing
           models will be deleted then.

           If None, the feature is disabled. It defines an upper bound on the
           models that can be used in the ensemble.

        seed: int = 1
            Seed used for the inidividual runs

        precision: 16 | 32 | 64 | 128 = 32
            Precision of floats to read the predictions

        memory_limit: int | None = None
            Memory limit in mb. If ``None``, no memory limit is enforced.

        read_at_most: int | None = None
            read at most n new prediction files in each iteration. If `None`, will read
            the predictions and calculate losses for all runs that require it.

        logger_port: int = DEFAULT_TCP_LOGGING_PORT
            Port that receives logging records

        start_time: float | None = None
            DISABLED: Just using time.time() to set it
            The time when this job was started, to account for any latency in job
            allocation.
        """
        self.time_left_for_ensembles = time_left_for_ensembles
        self.backend = backend
        self.dataset_name = dataset_name
        self.task = task
        self.metrics = metrics
        self.ensemble_class = ensemble_class
        self.ensemble_kwargs = ensemble_kwargs
        self.ensemble_nbest = ensemble_nbest
        self.max_models_on_disc = max_models_on_disc
        self.seed = seed
        self.precision = precision
        self.max_iterations = max_iterations
        self.read_at_most = read_at_most
        self.memory_limit = memory_limit
        self.random_state = check_random_state(random_state)
        self.logger_port = logger_port
        self.pynisher_context = pynisher_context

        # Store something similar to SMAC's runhistory
        self.history: list[dict[str, Any]] = []

        # We only submit new ensembles when there is not an active ensemble job
        self.futures: list[dask.distributed.Future] = []

        # The last criteria is the number of iterations
        self.iteration = 0

        # Keep track of when we started to know when we need to finish!
        self.start_time = time.time()

    def __call__(
        self,
        smbo: "SMBO",
        run_info: RunInfo,
        result: RunValue,
        time_left: float,
    ) -> None:
        """
        Returns
        -------
        List[Tuple[int, float, float, float]]:
            A list with the performance history of this ensemble, of the form
            [(pandas_timestamp, train_performance, val_performance, test_performance)]
        """
        if result.status in (StatusType.STOP, StatusType.ABORT) or smbo._stop:
            return
        self.build_ensemble(smbo.tae_runner.client)

    def build_ensemble(
        self,
        dask_client: dask.distributed.Client,
    ) -> None:
        """Build the ensemble

        Parameters
        ----------
        dask_client: dask.distributed.Client
            The dask client to use
        """
        # The second criteria is elapsed time
        elapsed_time = time.time() - self.start_time

        logger = get_named_client_logger(
            name="EnsembleBuilder",
            port=self.logger_port,
        )

        # First test for termination conditions
        if self.time_left_for_ensembles < elapsed_time:
            logger.info(
                "Terminate ensemble building as not time is left (run for {}s)".format(
                    elapsed_time
                ),
            )
            return
        if self.max_iterations is not None and self.max_iterations <= self.iteration:
            logger.info(
                "Terminate ensemble building because of max iterations:"
                f" {self.max_iterations} of {self.iteration}"
            )
            return

        if len(self.futures) != 0:
            if self.futures[0].done():
                result = self.futures.pop().result()
                if result:
                    ensemble_history, self.ensemble_nbest = result
                    logger.debug(
                        f"iteration={self.iteration} @ elapsed_time={elapsed_time}"
                        f" has history={ensemble_history}"
                    )
                    self.history.extend(ensemble_history)

        # Only submit new jobs if the previous ensemble job finished
        if len(self.futures) == 0:

            # Add the result of the run
            # On the next while iteration, no references to
            # ensemble builder object, so it should be garbage collected to
            # save memory while waiting for resources
            # Also, notice how ensemble nbest is returned, so we don't waste
            # iterations testing if the deterministic predictions size can
            # be fitted in memory
            try:
                # Submit a Dask job from this job, to properly
                # see it in the dask diagnostic dashboard
                # Notice that the forked ensemble_builder_process will
                # wait for the below function to be done
                self.futures.append(
                    dask_client.submit(
                        EnsembleBuilderManager.fit_and_return_ensemble,
                        backend=self.backend,
                        dataset_name=self.dataset_name,
                        task_type=self.task,
                        metrics=self.metrics,
                        ensemble_class=self.ensemble_class,
                        ensemble_kwargs=self.ensemble_kwargs,
                        ensemble_nbest=self.ensemble_nbest,
                        max_models_on_disc=self.max_models_on_disc,
                        seed=self.seed,
                        precision=self.precision,
                        memory_limit=self.memory_limit,
                        read_at_most=self.read_at_most,
                        random_state=self.random_state,
                        end_at=self.start_time + self.time_left_for_ensembles,
                        iteration=self.iteration,
                        pynisher_context=self.pynisher_context,
                        logger_port=self.logger_port,
                    )
                )

                logger.info(
                    "{}/{} Started Ensemble builder job at {} for iteration {}.".format(
                        # Log the client to make sure we
                        # remain connected to the scheduler
                        self.futures[0],
                        dask_client,
                        time.strftime("%Y.%m.%d-%H.%M.%S"),
                        self.iteration,
                    ),
                )
                self.iteration += 1
            except Exception as e:
                exception_traceback = traceback.format_exc()
                error_message = repr(e)
                logger.critical(exception_traceback)
                logger.critical(error_message)

    @staticmethod
    def fit_and_return_ensemble(
        iteration: int,
        end_at: float,
        backend: Backend,
        dataset_name: str,
        task_type: int,
        metrics: Sequence[Scorer],
        pynisher_context: str,
        ensemble_class: Type[AbstractEnsemble] = EnsembleSelection,
        ensemble_kwargs: Dict[str, Any] | None = None,
        ensemble_nbest: int | float = 50,
        max_models_on_disc: int | float | None = None,
        seed: int = 1,
        precision: int = 32,
        memory_limit: int | None = None,
        read_at_most: int | None = None,
        logger_port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
        random_state: int | np.random.RandomState | None = None,
    ) -> tuple[list[dict[str, Any]], int | float]:
        """
        A short function to fit and create an ensemble. It is just a wrapper to easily
        send a request to dask to create an ensemble and clean the memory when finished

        Parameters
        ----------
        iteration: int
            The current iteration

        end_at: float
            At what time the job must finish. Needs to be the endtime and not the
            time left because we do not know when dask schedules the job.

        backend: Backend
            Backend to write and read files

        dataset_name: str
            name of dataset

        task_type: int
            type of ML task

        metrics: Sequence[Scorer]
            Metrics to optimize the ensemble for.

        pynisher_context: "fork" | "spawn" | "forkserver" = "fork"
            Context to use for multiprocessing, can be either fork, spawn or forkserver.

        ensemble_class : Type[AbstractEnsemble] (default=EnsembleSelection)
            Class implementing the post-hoc ensemble algorithm. Set to
            ``None`` to disable ensemble building or use ``SingleBest``
            to obtain only use the single best model instead of an
            ensemble.

        ensemble_kwargs : Dict, optional
            Keyword arguments that are passed to the ensemble class upon
            initialization.

        ensemble_nbest: int | float = 50
            If int: consider only the n best prediction
            If float: consider only this fraction of the best models

        max_models_on_disc: int | float | None = 100
           Defines the maximum number of models that are kept in the disc.

           If int, it must be greater or equal than 1, and dictates the max number of
           models to keep.

           If float, it will be interpreted as the max megabytes allowed of disc space.
           That is, if the number of ensemble candidates require more disc space than
           this float value, the worst models will be deleted to keep within this
           budget. Models and predictions of the worst-performing models will be
           deleted then.

           If None, the feature is disabled.

        seed: int = 1
            Seed used for training the models in the backend

        precision: 16 | 32 | 64 | 128 = 32
            Precision of floats to read the predictions

        memory_limit: int | None = None
            Memory limit in mb. If ``None``, no memory limit is enforced.

        read_at_most: int | None = None
            read at most n new prediction files in each iteration. If `None`, will read
            the predictions and calculate losses for all runs that require it.

        logger_port: int = DEFAULT_TCP_LOGGING_PORT
            The port where the logging server is listening to.

        random_state: int | RandomState | None = None
            A random state used for the ensemble selection process.

        Returns
        -------
        (ensemble_history: list[dict[str, Any]], nbest: int | float)
            The ensemble history and the nbest chosen members
        """
        random_state = check_random_state(random_state)
        result = EnsembleBuilder(
            backend=backend,
            dataset_name=dataset_name,
            task_type=task_type,
            metrics=metrics,
            ensemble_class=ensemble_class,
            ensemble_kwargs=ensemble_kwargs,
            ensemble_nbest=ensemble_nbest,
            max_models_on_disc=max_models_on_disc,
            seed=seed,
            precision=precision,
            memory_limit=memory_limit,
            read_at_most=read_at_most,
            random_state=random_state.randint(10000000),
            logger_port=logger_port,
        ).run(
            end_at=end_at,
            iteration=iteration,
            pynisher_context=pynisher_context,
        )
        return result



from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from scipy.sparse import spmatrix

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class CategoricalImputation(AutoSklearnPreprocessingAlgorithm):
    """
    Substitute missing values by constant:
        When strategy == “constant”, fill_value is used to replace all
        occurrences of missing_values.
        If left to the default, fill_value will be 0 when imputing
        numerical data and “missing_value” for strings or object data types.
    """

    def __init__(
        self, random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "CategoricalImputation":
        import sklearn.impute

        if hasattr(X, "columns"):
            kind = X[X.columns[-1]].dtype.kind
        else:
            # Series, sparse and numpy have dtype
            # Only DataFrame does not
            kind = X.dtype.kind

        fill_value: Optional[int] = None

        number_kinds = ("i", "u", "f")
        if kind in number_kinds:
            if isinstance(X, spmatrix):
                # TODO negative labels
                #
                #   Previously this was the behaviour and went
                #   unnoticed. Imputing negative labels results in
                #   the cateogircal shift step failing as the ordinal
                #   encoder can't fix negative labels.
                #   This is here to document the behaviour explicitly
                fill_value = 0
            else:
                fill_value = min(np.unique(X)) - 1

        self.preprocessor = sklearn.impute.SimpleImputer(
            strategy="constant", copy=False, fill_value=fill_value
        )
        self.preprocessor.fit(X)
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        X = self.preprocessor.transform(X)
        return X

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "CategoricalImputation",
            "name": "Categorical Imputation",
            "handles_missing_values": True,
            "handles_nominal_values": True,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            # TODO find out of this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        return ConfigurationSpace()



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    UniformFloatHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool


class ARDRegression(AutoSklearnRegressionAlgorithm):
    def __init__(
        self,
        n_iter,
        tol,
        alpha_1,
        alpha_2,
        lambda_1,
        lambda_2,
        threshold_lambda,
        fit_intercept,
        random_state=None,
    ):
        self.random_state = random_state
        self.estimator = None

        self.n_iter = n_iter
        self.tol = tol
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.threshold_lambda = threshold_lambda
        self.fit_intercept = fit_intercept

    def fit(self, X, y):
        from sklearn.linear_model import ARDRegression

        self.n_iter = int(self.n_iter)
        self.tol = float(self.tol)
        self.alpha_1 = float(self.alpha_1)
        self.alpha_2 = float(self.alpha_2)
        self.lambda_1 = float(self.lambda_1)
        self.lambda_2 = float(self.lambda_2)
        self.threshold_lambda = float(self.threshold_lambda)
        self.fit_intercept = check_for_bool(self.fit_intercept)

        self.estimator = ARDRegression(
            n_iter=self.n_iter,
            tol=self.tol,
            alpha_1=self.alpha_1,
            alpha_2=self.alpha_2,
            lambda_1=self.lambda_1,
            lambda_2=self.lambda_2,
            compute_score=False,
            threshold_lambda=self.threshold_lambda,
            fit_intercept=True,
            normalize=False,
            copy_X=False,
            verbose=False,
        )

        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()

        self.estimator.fit(X, y)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "ARD",
            "name": "ARD Regression",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "prefers_data_normalized": True,
            "is_deterministic": True,
            "input": (DENSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()
        n_iter = UnParametrizedHyperparameter("n_iter", value=300)
        tol = UniformFloatHyperparameter(
            "tol", 10**-5, 10**-1, default_value=10**-3, log=True
        )
        alpha_1 = UniformFloatHyperparameter(
            name="alpha_1", lower=10**-10, upper=10**-3, default_value=10**-6
        )
        alpha_2 = UniformFloatHyperparameter(
            name="alpha_2",
            log=True,
            lower=10**-10,
            upper=10**-3,
            default_value=10**-6,
        )
        lambda_1 = UniformFloatHyperparameter(
            name="lambda_1",
            log=True,
            lower=10**-10,
            upper=10**-3,
            default_value=10**-6,
        )
        lambda_2 = UniformFloatHyperparameter(
            name="lambda_2",
            log=True,
            lower=10**-10,
            upper=10**-3,
            default_value=10**-6,
        )
        threshold_lambda = UniformFloatHyperparameter(
            name="threshold_lambda",
            log=True,
            lower=10**3,
            upper=10**5,
            default_value=10**4,
        )
        fit_intercept = UnParametrizedHyperparameter("fit_intercept", "True")

        cs.add_hyperparameters(
            [
                n_iter,
                tol,
                alpha_1,
                alpha_2,
                lambda_1,
                lambda_2,
                threshold_lambda,
                fit_intercept,
            ]
        )

        return cs



from __future__ import annotations

from typing import Sequence

import warnings

import numpy as np
from sklearn.exceptions import NotFittedError

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import TASK_TYPES
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES
from autosklearn.ensemble_building.run import Run
from autosklearn.ensembles.abstract_ensemble import (
    AbstractEnsemble,
    AbstractMultiObjectiveEnsemble,
)
from autosklearn.ensembles.singlebest_ensemble import SingleModelEnsemble
from autosklearn.metrics import Scorer, calculate_losses
from autosklearn.pipeline.base import BasePipeline
from autosklearn.util.multiobjective import pareto_front


class MultiObjectiveDummyEnsemble(AbstractMultiObjectiveEnsemble):
    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        random_state: int | np.random.RandomState | None = None,
    ) -> None:
        """A dummy implementation of a multi-objective ensemble.

        Builds ensembles that are individual models on the Pareto front each.

        Parameters
        ----------
        task_type: int
            An identifier indicating which task is being performed.

        metrics: Sequence[Scorer] | Scorer
            The metrics used to evaluate the models.

        backend : Backend
            Gives access to the backend of Auto-sklearn. Not used.

        random_state: int | RandomState | None = None
            Not used.
        """
        self.task_type = task_type
        if isinstance(metrics, Sequence):
            if len(metrics) == 1:
                warnings.warn(
                    "Passed only a single metric to a multi-objective ensemble. "
                    "Please use a single-objective ensemble in such cases."
                )
            self.metrics = metrics
        else:
            self.metric = [metrics]
        self.random_state = random_state
        self.backend = backend

    @property
    def pareto_set(self) -> Sequence[AbstractEnsemble]:
        if not hasattr(self, "pareto_set_"):
            raise NotFittedError("`pareto_set` not created, please call `fit()` first")
        return self.pareto_set_

    def fit(
        self,
        base_models_predictions: list[np.ndarray],
        true_targets: np.ndarray,
        model_identifiers: list[tuple[int, int, float]],
        runs: Sequence[Run],
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> MultiObjectiveDummyEnsemble:
        """Select dummy ensembles given predictions of base models and targets.

        Parameters
        ----------
        base_models_predictions: np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

            Can be a list of 2d numpy arrays as well to prevent copying all
            predictions into a single, large numpy array.

        true_targets : array of shape [n_targets]

        model_identifiers : identifier for each base model.
            Can be used for practical text output of the ensemble.

        runs: Sequence[Run]
            Additional information for each run executed by SMAC that was
            considered by the ensemble builder. Not used.

        X_data : list-like | sparse matrix | None = None
            X data to give to the metric if required

        Returns
        -------
        self
        """
        if self.task_type not in TASK_TYPES:
            raise ValueError("Unknown task type %s." % self.task_type)

        all_costs = np.empty((len(base_models_predictions), len(self.metrics)))
        for i, base_model_prediction in enumerate(base_models_predictions):
            losses = calculate_losses(
                solution=true_targets,
                prediction=base_model_prediction,
                task_type=self.task_type,
                metrics=self.metrics,
                X_data=X_data,
            )
            all_costs[i] = [losses[metric.name] for metric in self.metrics]

        all_costs = np.array(all_costs)
        sort_by_first_metric = np.argsort(all_costs[:, 0])
        efficient_points = pareto_front(all_costs, is_loss=True)
        pareto_set = []

        for argsort_idx in sort_by_first_metric:
            if not efficient_points[argsort_idx]:
                continue
            ensemble = SingleModelEnsemble(
                task_type=self.task_type,
                metrics=self.metrics,
                random_state=self.random_state,
                backend=self.backend,
                model_index=argsort_idx,
            )
            ensemble.fit(
                base_models_predictions=base_models_predictions,
                true_targets=true_targets,
                model_identifiers=model_identifiers,
                runs=runs,
                X_data=X_data,
            )
            pareto_set.append(ensemble)

        self.pareto_set_ = pareto_set
        return self

    def predict(
        self, base_models_predictions: np.ndarray | list[np.ndarray]
    ) -> np.ndarray:
        """Predict using the ensemble which is best for the 1st metric.

        Parameters
        ----------
        base_models_predictions : np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            Same as in the fit method.

        Returns
        -------
        np.ndarray
        """
        return self.pareto_set[0].predict(base_models_predictions)

    def __str__(self) -> str:
        return "MultiObjectiveDummyEnsemble: %d models" % len(self.pareto_set)

    def get_models_with_weights(
        self, models: dict[tuple[int, int, float], BasePipeline]
    ) -> list[tuple[float, BasePipeline]]:
        """Return a list of (weight, model) pairs for the ensemble that is
        best for the 1st metric.

        Parameters
        ----------
        models : dict {identifier : model object}
            The identifiers are the same as the one presented to the fit()
            method. Models can be used for nice printing.

        Returns
        -------
        list[tuple[float, BasePipeline]]
        """
        return self.pareto_set[0].get_models_with_weights(models)

    def get_identifiers_with_weights(
        self,
    ) -> list[tuple[tuple[int, int, float], float]]:
        """Return a (identifier, weight)-pairs for all models that were passed to the
        ensemble builder based on the ensemble that is best for the 1st metric.

        Parameters
        ----------
        models : dict {identifier : model object}
            The identifiers are the same as the one presented to the fit()
            method. Models can be used for nice printing.

        Returns
        -------
        list[tuple[tuple[int, int, float], float]
        """
        return self.pareto_set[0].get_identifiers_with_weights()

    def get_selected_model_identifiers(self) -> list[tuple[int, int, float]]:
        """Return identifiers of models in the ensemble that is best for the 1st metric.

        This includes models which have a weight of zero!

        Returns
        -------
        list
        """
        return self.pareto_set[0].get_selected_model_identifiers()

    def get_validation_performance(self) -> float:
        """Validation performance of the ensemble that is best for the 1st metric.

        Returns
        -------
        float
        """
        return self.pareto_set[0].get_validation_performance()



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.pipeline.implementations.util import (
    convert_multioutput_multiclass_to_multilabel,
)
from autosklearn.util.common import check_none


class DecisionTree(AutoSklearnClassificationAlgorithm):
    def __init__(
        self,
        criterion,
        max_features,
        max_depth_factor,
        min_samples_split,
        min_samples_leaf,
        min_weight_fraction_leaf,
        max_leaf_nodes,
        min_impurity_decrease,
        class_weight=None,
        random_state=None,
    ):
        self.criterion = criterion
        self.max_features = max_features
        self.max_depth_factor = max_depth_factor
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.random_state = random_state
        self.class_weight = class_weight
        self.estimator = None

    def fit(self, X, y, sample_weight=None):
        from sklearn.tree import DecisionTreeClassifier

        self.max_features = float(self.max_features)
        # Heuristic to set the tree depth
        if check_none(self.max_depth_factor):
            max_depth_factor = self.max_depth_factor = None
        else:
            num_features = X.shape[1]
            self.max_depth_factor = int(self.max_depth_factor)
            max_depth_factor = max(
                1, int(np.round(self.max_depth_factor * num_features, 0))
            )
        self.min_samples_split = int(self.min_samples_split)
        self.min_samples_leaf = int(self.min_samples_leaf)
        if check_none(self.max_leaf_nodes):
            self.max_leaf_nodes = None
        else:
            self.max_leaf_nodes = int(self.max_leaf_nodes)
        self.min_weight_fraction_leaf = float(self.min_weight_fraction_leaf)
        self.min_impurity_decrease = float(self.min_impurity_decrease)

        self.estimator = DecisionTreeClassifier(
            criterion=self.criterion,
            max_depth=max_depth_factor,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            max_leaf_nodes=self.max_leaf_nodes,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
            class_weight=self.class_weight,
            random_state=self.random_state,
        )
        self.estimator.fit(X, y, sample_weight=sample_weight)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        probas = self.estimator.predict_proba(X)
        probas = convert_multioutput_multiclass_to_multilabel(probas)
        return probas

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "DT",
            "name": "Decision Tree Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        criterion = CategoricalHyperparameter(
            "criterion", ["gini", "entropy"], default_value="gini"
        )
        max_depth_factor = UniformFloatHyperparameter(
            "max_depth_factor", 0.0, 2.0, default_value=0.5
        )
        min_samples_split = UniformIntegerHyperparameter(
            "min_samples_split", 2, 20, default_value=2
        )
        min_samples_leaf = UniformIntegerHyperparameter(
            "min_samples_leaf", 1, 20, default_value=1
        )
        min_weight_fraction_leaf = Constant("min_weight_fraction_leaf", 0.0)
        max_features = UnParametrizedHyperparameter("max_features", 1.0)
        max_leaf_nodes = UnParametrizedHyperparameter("max_leaf_nodes", "None")
        min_impurity_decrease = UnParametrizedHyperparameter(
            "min_impurity_decrease", 0.0
        )

        cs.add_hyperparameters(
            [
                criterion,
                max_features,
                max_depth_factor,
                min_samples_split,
                min_samples_leaf,
                min_weight_fraction_leaf,
                max_leaf_nodes,
                min_impurity_decrease,
            ]
        )

        return cs



import typing
from typing import Dict, List, Optional, Sequence

import copy
import json
import logging
import multiprocessing
import os
import time
import traceback
import warnings

import dask.distributed
import pynisher
from smac.callbacks import IncorporateRunResultCallback
from smac.facade.smac_ac_facade import SMAC4AC
from smac.intensification.intensification import Intensifier
from smac.intensification.simple_intensifier import SimpleIntensifier
from smac.optimizer.multi_objective.parego import ParEGO
from smac.runhistory.runhistory2epm import RunHistory2EPM4LogCost
from smac.scenario.scenario import Scenario
from smac.tae.dask_runner import DaskParallelRunner
from smac.tae.serial_runner import SerialRunner

import autosklearn.metalearning
from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    CLASSIFICATION_TASKS,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
    TASK_TYPES_TO_STRING,
)
from autosklearn.data.abstract_data_manager import AbstractDataManager
from autosklearn.ensemble_building import EnsembleBuilderManager
from autosklearn.evaluation import ExecuteTaFuncWithQueue, get_cost_of_crash
from autosklearn.metalearning.metafeatures.metafeatures import (
    calculate_all_metafeatures_encoded_labels,
    calculate_all_metafeatures_with_labels,
)
from autosklearn.metalearning.metalearning.meta_base import MetaBase
from autosklearn.metalearning.mismbo import suggest_via_metalearning
from autosklearn.metrics import Scorer
from autosklearn.util.logging_ import get_named_client_logger
from autosklearn.util.parallel import preload_modules
from autosklearn.util.stopwatch import StopWatch

EXCLUDE_META_FEATURES_CLASSIFICATION = {
    "Landmark1NN",
    "LandmarkDecisionNodeLearner",
    "LandmarkDecisionTree",
    "LandmarkLDA",
    "LandmarkNaiveBayes",
    "LandmarkRandomNodeLearner",
    "PCAFractionOfComponentsFor95PercentVariance",
    "PCAKurtosisFirstPC",
    "PCASkewnessFirstPC",
    "PCA",
}

EXCLUDE_META_FEATURES_REGRESSION = {
    "Landmark1NN",
    "LandmarkDecisionNodeLearner",
    "LandmarkDecisionTree",
    "LandmarkLDA",
    "LandmarkNaiveBayes",
    "PCAFractionOfComponentsFor95PercentVariance",
    "PCAKurtosisFirstPC",
    "PCASkewnessFirstPC",
    "NumberOfClasses",
    "ClassOccurences",
    "ClassProbabilityMin",
    "ClassProbabilityMax",
    "ClassProbabilityMean",
    "ClassProbabilitySTD",
    "ClassEntropy",
    "LandmarkRandomNodeLearner",
    "PCA",
}


def get_send_warnings_to_logger(logger):
    def _send_warnings_to_log(message, category, filename, lineno, file, line):
        logger.debug("%s:%s: %s:%s", filename, lineno, category.__name__, message)

    return _send_warnings_to_log


# metalearning helpers
def _calculate_metafeatures(
    data_feat_type,
    data_info_task,
    basename,
    x_train,
    y_train,
    stopwatch: StopWatch,
    logger_,
):
    with warnings.catch_warnings():
        warnings.showwarning = get_send_warnings_to_logger(logger_)

        # == Calculate metafeatures
        with stopwatch.time("Calculate meta-features") as task_timer:

            EXCLUDE_META_FEATURES = (
                EXCLUDE_META_FEATURES_CLASSIFICATION
                if data_info_task in CLASSIFICATION_TASKS
                else EXCLUDE_META_FEATURES_REGRESSION
            )

            if data_info_task in [
                MULTICLASS_CLASSIFICATION,
                BINARY_CLASSIFICATION,
                MULTILABEL_CLASSIFICATION,
                REGRESSION,
                MULTIOUTPUT_REGRESSION,
            ]:
                logger_.info("Start calculating metafeatures for %s", basename)
                result = calculate_all_metafeatures_with_labels(
                    x_train,
                    y_train,
                    feat_type=data_feat_type,
                    dataset_name=basename,
                    dont_calculate=EXCLUDE_META_FEATURES,
                    logger=logger_,
                )
                for key in list(result.metafeature_values.keys()):
                    if result.metafeature_values[key].type_ != "METAFEATURE":
                        del result.metafeature_values[key]

            else:
                result = None
                logger_.info("Metafeatures not calculated")

        logger_.info(f"{task_timer.name} took {task_timer.wall_duration:5.2f}"),
        return result


def _calculate_metafeatures_encoded(
    data_feat_type,
    basename,
    x_train,
    y_train,
    stopwatch: StopWatch,
    task,
    logger_,
):
    with warnings.catch_warnings():
        warnings.showwarning = get_send_warnings_to_logger(logger_)

        EXCLUDE_META_FEATURES = (
            EXCLUDE_META_FEATURES_CLASSIFICATION
            if task in CLASSIFICATION_TASKS
            else EXCLUDE_META_FEATURES_REGRESSION
        )

        with stopwatch.time("Calculate meta-features encoded") as task_timer:

            result = calculate_all_metafeatures_encoded_labels(
                x_train,
                y_train,
                feat_type=data_feat_type,
                dataset_name=basename,
                dont_calculate=EXCLUDE_META_FEATURES,
                logger=logger_,
            )
            for key in list(result.metafeature_values.keys()):
                if result.metafeature_values[key].type_ != "METAFEATURE":
                    del result.metafeature_values[key]

        logger_.info(f"{task_timer.name} took {task_timer.wall_duration:5.2f}sec")
        return result


def _get_metalearning_configurations(
    meta_base,
    basename,
    metric,
    configuration_space,
    task,
    initial_configurations_via_metalearning,
    stopwatch: StopWatch,
    is_sparse,
    logger,
):
    try:
        metalearning_configurations = suggest_via_metalearning(
            meta_base,
            basename,
            metric,
            task,
            is_sparse == 1,
            initial_configurations_via_metalearning,
            logger=logger,
        )
    except Exception as e:
        logger.error("Error getting metalearning configurations!")
        logger.error(str(e))
        logger.error(traceback.format_exc())
        metalearning_configurations = []

    return metalearning_configurations


def get_smac_object(
    scenario_dict,
    seed,
    ta,
    ta_kwargs,
    metalearning_configurations,
    n_jobs,
    dask_client,
    multi_objective_algorithm,
    multi_objective_kwargs,
):
    if len(scenario_dict["instances"]) > 1:
        intensifier = Intensifier
    else:
        intensifier = SimpleIntensifier

    scenario = Scenario(scenario_dict)
    if len(metalearning_configurations) > 0:
        default_config = scenario.cs.get_default_configuration()
        initial_configurations = [default_config] + metalearning_configurations
    else:
        initial_configurations = None
    rh2EPM = RunHistory2EPM4LogCost
    return SMAC4AC(
        scenario=scenario,
        rng=seed,
        runhistory2epm=rh2EPM,
        tae_runner=ta,
        tae_runner_kwargs=ta_kwargs,
        initial_configurations=initial_configurations,
        run_id=seed,
        intensifier=intensifier,
        dask_client=dask_client,
        n_jobs=n_jobs,
        multi_objective_algorithm=multi_objective_algorithm,
        multi_objective_kwargs=multi_objective_kwargs,
    )


class AutoMLSMBO:
    def __init__(
        self,
        config_space,
        dataset_name,
        backend,
        total_walltime_limit,
        func_eval_time_limit,
        memory_limit,
        metrics: Sequence[Scorer],
        stopwatch: StopWatch,
        n_jobs,
        dask_client: dask.distributed.Client,
        port: int,
        start_num_run=1,
        data_memory_limit=None,
        num_metalearning_cfgs=25,
        config_file=None,
        seed=1,
        metadata_directory=None,
        resampling_strategy="holdout",
        resampling_strategy_args=None,
        include: Optional[Dict[str, List[str]]] = None,
        exclude: Optional[Dict[str, List[str]]] = None,
        disable_file_output=False,
        smac_scenario_args=None,
        get_smac_object_callback=None,
        scoring_functions=None,
        pynisher_context="spawn",
        ensemble_callback: typing.Optional[EnsembleBuilderManager] = None,
        trials_callback: typing.Optional[IncorporateRunResultCallback] = None,
    ):
        super(AutoMLSMBO, self).__init__()
        # data related
        self.dataset_name = dataset_name
        self.datamanager = None
        self.metrics = metrics
        self.task = None
        self.backend = backend
        self.port = port

        # the configuration space
        self.config_space = config_space

        # the number of parallel workers/jobs
        self.n_jobs = n_jobs
        self.dask_client = dask_client

        # Evaluation
        self.resampling_strategy = resampling_strategy
        if resampling_strategy_args is None:
            resampling_strategy_args = {}
        self.resampling_strategy_args = resampling_strategy_args

        # and a bunch of useful limits
        self.worst_possible_result = get_cost_of_crash(self.metrics)
        self.total_walltime_limit = int(total_walltime_limit)
        self.func_eval_time_limit = int(func_eval_time_limit)
        self.memory_limit = memory_limit
        self.data_memory_limit = data_memory_limit
        self.stopwatch = stopwatch
        self.num_metalearning_cfgs = num_metalearning_cfgs
        self.config_file = config_file
        self.seed = seed
        self.metadata_directory = metadata_directory
        self.start_num_run = start_num_run
        self.include = include
        self.exclude = exclude
        self.disable_file_output = disable_file_output
        self.smac_scenario_args = smac_scenario_args
        self.get_smac_object_callback = get_smac_object_callback
        self.scoring_functions = scoring_functions

        self.pynisher_context = pynisher_context

        self.ensemble_callback = ensemble_callback
        self.trials_callback = trials_callback

        dataset_name_ = "" if dataset_name is None else dataset_name
        logger_name = "%s(%d):%s" % (
            self.__class__.__name__,
            self.seed,
            ":" + dataset_name_,
        )
        if port is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = get_named_client_logger(
                name=logger_name,
                port=self.port,
            )

    def reset_data_manager(self, max_mem=None):
        if max_mem is None:
            max_mem = self.data_memory_limit
        if self.datamanager is not None:
            del self.datamanager
        if isinstance(self.dataset_name, AbstractDataManager):
            self.datamanager = self.dataset_name
        else:
            self.datamanager = self.backend.load_datamanager()

        self.task = self.datamanager.info["task"]

    def collect_metalearning_suggestions(self, meta_base):

        with self.stopwatch.time("Initial Configurations") as task:
            metalearning_configurations = _get_metalearning_configurations(
                meta_base=meta_base,
                basename=self.dataset_name,
                metric=self.metrics[0],
                configuration_space=self.config_space,
                task=self.task,
                is_sparse=self.datamanager.info["is_sparse"],
                initial_configurations_via_metalearning=self.num_metalearning_cfgs,
                stopwatch=self.stopwatch,
                logger=self.logger,
            )

        self.logger.debug(f"Initial Configurations: {len(metalearning_configurations)}")
        for config in metalearning_configurations:
            self.logger.debug(config)

        self.logger.debug(f"{task.name} took {task.wall_duration:5.2f}sec")

        time_since_start = self.stopwatch.time_since(self.dataset_name, "start")
        time_left = self.total_walltime_limit - time_since_start
        self.logger.info(f"Time left for {task.name}: {time_left:5.2f}s")

        return metalearning_configurations

    def _calculate_metafeatures_with_limits(self, time_limit):
        res = None
        time_limit = max(time_limit, 1)
        try:
            context = multiprocessing.get_context(self.pynisher_context)
            preload_modules(context)
            safe_mf = pynisher.enforce_limits(
                mem_in_mb=self.memory_limit,
                wall_time_in_s=int(time_limit),
                grace_period_in_s=30,
                context=context,
                logger=self.logger,
            )(_calculate_metafeatures)
            res = safe_mf(
                data_feat_type=self.datamanager.feat_type,
                data_info_task=self.datamanager.info["task"],
                x_train=self.datamanager.data["X_train"],
                y_train=self.datamanager.data["Y_train"],
                basename=self.dataset_name,
                stopwatch=self.stopwatch,
                logger_=self.logger,
            )
        except Exception as e:
            self.logger.error("Error getting metafeatures: %s", str(e))

        return res

    def _calculate_metafeatures_encoded_with_limits(self, time_limit):
        res = None
        time_limit = max(time_limit, 1)
        try:
            context = multiprocessing.get_context(self.pynisher_context)
            preload_modules(context)
            safe_mf = pynisher.enforce_limits(
                mem_in_mb=self.memory_limit,
                wall_time_in_s=int(time_limit),
                grace_period_in_s=30,
                context=context,
                logger=self.logger,
            )(_calculate_metafeatures_encoded)
            res = safe_mf(
                data_feat_type=self.datamanager.feat_type,
                task=self.datamanager.info["task"],
                x_train=self.datamanager.data["X_train"],
                y_train=self.datamanager.data["Y_train"],
                basename=self.dataset_name,
                stopwatch=self.stopwatch,
                logger_=self.logger,
            )
        except Exception as e:
            self.logger.error("Error getting metafeatures (encoded) : %s", str(e))

        return res

    def run_smbo(self):

        self.stopwatch.start("SMBO")

        # == first things first: load the datamanager
        self.reset_data_manager()

        # == Initialize non-SMBO stuff
        # first create a scenario
        seed = self.seed
        self.config_space.seed(seed)
        # allocate a run history
        num_run = self.start_num_run

        # Initialize some SMAC dependencies

        metalearning_configurations = self.get_metalearning_suggestions()

        if self.resampling_strategy in ["partial-cv", "partial-cv-iterative-fit"]:
            num_folds = self.resampling_strategy_args["folds"]
            instances = [
                [json.dumps({"task_id": self.dataset_name, "fold": fold_number})]
                for fold_number in range(num_folds)
            ]
        else:
            instances = [[json.dumps({"task_id": self.dataset_name})]]

        # TODO rebuild target algorithm to be it's own target algorithm
        # evaluator, which takes into account that a run can be killed prior
        # to the model being fully fitted; thus putting intermediate results
        # into a queue and querying them once the time is over

        ta_kwargs = dict(
            backend=copy.deepcopy(self.backend),
            autosklearn_seed=seed,
            resampling_strategy=self.resampling_strategy,
            initial_num_run=num_run,
            include=self.include,
            exclude=self.exclude,
            metrics=self.metrics,
            memory_limit=self.memory_limit,
            disable_file_output=self.disable_file_output,
            scoring_functions=self.scoring_functions,
            port=self.port,
            pynisher_context=self.pynisher_context,
            **self.resampling_strategy_args,
        )
        ta = ExecuteTaFuncWithQueue

        startup_time = self.stopwatch.time_since(self.dataset_name, "start")
        total_walltime_limit = self.total_walltime_limit - startup_time - 5

        scenario_dict = {
            "abort_on_first_run_crash": False,
            "save-results-instantly": True,
            "cs": self.config_space,
            "cutoff_time": self.func_eval_time_limit,
            "deterministic": "true",
            "instances": instances,
            "memory_limit": self.memory_limit,
            "output-dir": self.backend.get_smac_output_directory(),
            "run_obj": "quality",
            "wallclock_limit": total_walltime_limit,
            "cost_for_crash": self.worst_possible_result,
        }
        if self.smac_scenario_args is not None:
            for arg in [
                "abort_on_first_run_crash",
                "cs",
                "deterministic",
                "instances",
                "output-dir",
                "run_obj",
                "shared-model",
                "cost_for_crash",
            ]:
                if arg in self.smac_scenario_args:
                    self.logger.warning(
                        "Cannot override scenario argument %s, " "will ignore this.",
                        arg,
                    )
                    del self.smac_scenario_args[arg]
            for arg in [
                "cutoff_time",
                "memory_limit",
                "wallclock_limit",
            ]:
                if arg in self.smac_scenario_args:
                    self.logger.warning(
                        "Overriding scenario argument %s: %s with value %s",
                        arg,
                        scenario_dict[arg],
                        self.smac_scenario_args[arg],
                    )
            scenario_dict.update(self.smac_scenario_args)

        smac_args = {
            "scenario_dict": scenario_dict,
            "seed": seed,
            "ta": ta,
            "ta_kwargs": ta_kwargs,
            "metalearning_configurations": metalearning_configurations,
            "n_jobs": self.n_jobs,
            "dask_client": self.dask_client,
        }
        if len(self.metrics) > 1:
            smac_args["multi_objective_algorithm"] = ParEGO
            smac_args["multi_objective_kwargs"] = {"rho": 0.05}
            scenario_dict["multi_objectives"] = [metric.name for metric in self.metrics]
        else:
            smac_args["multi_objective_algorithm"] = None
            smac_args["multi_objective_kwargs"] = {}
        if self.get_smac_object_callback is not None:
            smac = self.get_smac_object_callback(**smac_args)
        else:
            smac = get_smac_object(**smac_args)

        if self.ensemble_callback is not None:
            smac.register_callback(self.ensemble_callback)
        if self.trials_callback is not None:
            smac.register_callback(self.trials_callback)

        smac.optimize()

        self.runhistory = smac.solver.runhistory
        self.trajectory = smac.solver.intensifier.traj_logger.trajectory
        if isinstance(smac.solver.tae_runner, DaskParallelRunner):
            self._budget_type = smac.solver.tae_runner.single_worker.budget_type
        elif isinstance(smac.solver.tae_runner, SerialRunner):
            self._budget_type = smac.solver.tae_runner.budget_type
        else:
            raise NotImplementedError(type(smac.solver.tae_runner))

        self.stopwatch.stop("SMBO")

        return self.runhistory, self.trajectory, self._budget_type

    def get_metalearning_suggestions(self):
        # == METALEARNING suggestions
        # we start by evaluating the defaults on the full dataset again
        # and add the suggestions from metalearning behind it
        if self.num_metalearning_cfgs > 0:
            # If metadata directory is None, use default
            if self.metadata_directory is None:
                metalearning_directory = os.path.dirname(
                    autosklearn.metalearning.__file__
                )
                # There is no multilabel data in OpenML
                if self.task == MULTILABEL_CLASSIFICATION:
                    meta_task = BINARY_CLASSIFICATION
                else:
                    meta_task = self.task
                metadata_directory = os.path.join(
                    metalearning_directory,
                    "files",
                    "%s_%s_%s"
                    % (
                        self.metrics[0],
                        TASK_TYPES_TO_STRING[meta_task],
                        "sparse" if self.datamanager.info["is_sparse"] else "dense",
                    ),
                )
                self.metadata_directory = metadata_directory

            # If metadata directory is specified by user,
            # then verify that it exists.
            else:
                if not os.path.exists(self.metadata_directory):
                    raise ValueError(
                        "The specified metadata directory '%s' "
                        "does not exist!" % self.metadata_directory
                    )

                else:
                    # There is no multilabel data in OpenML
                    if self.task == MULTILABEL_CLASSIFICATION:
                        meta_task = BINARY_CLASSIFICATION
                    else:
                        meta_task = self.task

                    metadata_directory = os.path.join(
                        self.metadata_directory,
                        "%s_%s_%s"
                        % (
                            self.metrics[0],
                            TASK_TYPES_TO_STRING[meta_task],
                            "sparse" if self.datamanager.info["is_sparse"] else "dense",
                        ),
                    )
                    # Check that the metadata directory has the correct
                    # subdirectory needed for this dataset.
                    if os.path.basename(metadata_directory) not in os.listdir(
                        self.metadata_directory
                    ):
                        raise ValueError(
                            "The specified metadata directory "
                            "'%s' does not have the correct "
                            "subdirectory '%s'"
                            % (
                                self.metadata_directory,
                                os.path.basename(metadata_directory),
                            )
                        )
                self.metadata_directory = metadata_directory

            if os.path.exists(self.metadata_directory):

                self.logger.info("Metadata directory: %s", self.metadata_directory)
                meta_base = MetaBase(
                    self.config_space, self.metadata_directory, self.logger
                )

                metafeature_calculation_time_limit = int(self.total_walltime_limit / 4)
                metafeature_calculation_start_time = time.time()
                meta_features = self._calculate_metafeatures_with_limits(
                    metafeature_calculation_time_limit
                )
                metafeature_calculation_end_time = time.time()
                metafeature_calculation_time_limit = (
                    metafeature_calculation_time_limit
                    - (
                        metafeature_calculation_end_time
                        - metafeature_calculation_start_time
                    )
                )

                if metafeature_calculation_time_limit < 1:
                    self.logger.warning(
                        "Time limit for metafeature calculation less "
                        "than 1 seconds (%f). Skipping calculation "
                        "of metafeatures for encoded dataset.",
                        metafeature_calculation_time_limit,
                    )
                    meta_features_encoded = None
                else:
                    with warnings.catch_warnings():
                        warnings.showwarning = get_send_warnings_to_logger(self.logger)
                    meta_features_encoded = (
                        self._calculate_metafeatures_encoded_with_limits(
                            metafeature_calculation_time_limit
                        )
                    )

                # In case there is a problem calculating the encoded meta-features
                if meta_features is None:
                    if meta_features_encoded is not None:
                        meta_features = meta_features_encoded
                else:
                    if meta_features_encoded is not None:
                        meta_features.metafeature_values.update(
                            meta_features_encoded.metafeature_values
                        )

                if meta_features is not None:
                    meta_base.add_dataset(self.dataset_name, meta_features)
                    # Do mean imputation of the meta-features - should be done specific
                    # for each prediction model!
                    all_metafeatures = meta_base.get_metafeatures(
                        features=list(meta_features.keys())
                    )
                    all_metafeatures.fillna(all_metafeatures.mean(), inplace=True)

                    with warnings.catch_warnings():
                        warnings.showwarning = get_send_warnings_to_logger(self.logger)
                        metalearning_configurations = (
                            self.collect_metalearning_suggestions(meta_base)
                        )
                    if metalearning_configurations is None:
                        metalearning_configurations = []
                    self.reset_data_manager()

                    self.logger.info("%s", meta_features)

                    # Convert meta-features into a dictionary because the scenario
                    # expects a dictionary
                    meta_features_dict = {}
                    for dataset, series in all_metafeatures.iterrows():
                        meta_features_dict[dataset] = series.values
                    meta_features_list = []
                    for meta_feature_name in all_metafeatures.columns:
                        meta_features_list.append(
                            meta_features[meta_feature_name].value
                        )
                    self.logger.info(list(meta_features_dict.keys()))

            else:
                meta_features = None
                self.logger.warning(
                    "Could not find meta-data directory %s" % metadata_directory
                )

        else:
            meta_features = None
        if meta_features is None:
            metalearning_configurations = []
        return metalearning_configurations



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class Densifier(AutoSklearnPreprocessingAlgorithm):
    def __init__(self, random_state=None):
        pass

    def fit(self, X, y=None):
        self.fitted_ = True
        return self

    def transform(self, X):
        from scipy import sparse

        if sparse.issparse(X):
            return X.todense().getA()
        else:
            return X

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "RandomTreesEmbedding",
            "name": "Random Trees Embedding",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (SPARSE, UNSIGNED_DATA),
            "output": (DENSE, INPUT),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()
        return cs



"""Test the _model_predict helper function such that it shapes output correctly"""
from typing import Callable, Dict, Tuple

import warnings

import numpy as np
from sklearn.ensemble import VotingClassifier, VotingRegressor

from autosklearn.automl import _model_predict
from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
)
from autosklearn.util.logging_ import PicklableClientLogger

from pytest_cases import parametrize


class WarningModel:
    """Simple model that returns incorrect shape and issues warning"""

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Shout a warning during prediction"""
        warnings.warn("shout")
        return X


@parametrize(
    "dataspec, expected_shape",
    [
        ({"kind": BINARY_CLASSIFICATION, "dims": (100, 5)}, (100, 2)),
        ({"kind": MULTICLASS_CLASSIFICATION, "dims": (100, 5), "classes": 3}, (100, 3)),
        (
            {
                "kind": MULTILABEL_CLASSIFICATION,
                "dims": (100, 5),
                "classes": [[0, 0], [0, 1], [1, 0], [1, 1]],
            },
            (100, 2),  # TODO seems wrong
        ),
    ],
)
def test_classifier_output_shape(
    dataspec: Dict,
    expected_shape: Tuple[int, ...],
    make_voting_classifier: Callable[..., VotingClassifier],
    make_data: Callable[..., Tuple[np.ndarray, np.ndarray]],
) -> None:
    """
    Parameters
    ----------
    dataspec : Dict
        The spec to make data of

    expected_shape : Tuple[int, ...]
        The expected shape of the output of _model_predict

    Fixtures
    --------
    make_voting_classifier : Callable[..., VotingClassifier]
        Factory to make a voting classifier which _model_predict expects

    make_data : Callable[..., Tuple[np.ndarray, np.ndarray]]
        Factory to make data according to a spec

    Expects
    -------
    * The output shape after predicting should be the expected shape

    Note
    ----
    * The output shape for MULTILABEL_CLASSIFICATION seems wrong according to

    """
    task = dataspec["kind"]
    X, y = make_data(**dataspec)

    voter = make_voting_classifier(X=X, y=y)

    output = _model_predict(voter, X, task=task)
    assert output.shape == expected_shape


@parametrize(
    "dataspec, expected_shape",
    [
        ({"kind": REGRESSION, "dims": (100, 5)}, (100,)),
        ({"kind": MULTIOUTPUT_REGRESSION, "dims": (100, 5), "targets": 3}, (100, 3)),
    ],
)
def test_regressor_output_shape(
    dataspec: Dict,
    expected_shape: Tuple[int, ...],
    make_voting_regressor: Callable[..., VotingRegressor],
    make_data: Callable[..., Tuple[np.ndarray, np.ndarray]],
) -> None:
    """
    Parameters
    ----------
    dataspec : Dict
        The spec to make data of

    expected_shape : Tuple[int, ...]
        The expected shape of the output of _model_predict

    Fixtures
    --------
    make_voting_regressor: Callable[..., VotingRegressor]
        Factory to make a voting classifier which _model_predict expects

    make_data : Callable[..., Tuple[np.ndarray, np.ndarray]]
        Factory to make data according to a spec
    """
    task = dataspec["kind"]
    X, y = make_data(**dataspec)

    voter = make_voting_regressor(X=X, y=y)

    output = _model_predict(voter, X, task=task)
    assert output.shape == expected_shape


def test_outputs_warnings_to_logs(
    mock_logger: PicklableClientLogger,
) -> None:
    """
    Fixtures
    --------
    mock_logger : PicklableClientLogger
        A mock logger that can be queried for call counts

    Expects
    -------
    * Any warning emitted by a model should be redirected to the logger
    """
    _model_predict(
        model=WarningModel(), X=np.eye(5), task=REGRESSION, logger=mock_logger
    )

    assert mock_logger.warning.call_count == 1  # type: ignore


def test_outputs_to_stdout_if_no_logger() -> None:
    """
    Expects
    -------
    * With no logger, any warning emitted by a model goes to standard out
    """
    with warnings.catch_warnings(record=True) as w:
        _model_predict(model=WarningModel(), X=np.eye(5), task=REGRESSION, logger=None)

    assert len(w) == 1, "One warning sould have been emmited"



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool


class PCA(AutoSklearnPreprocessingAlgorithm):
    def __init__(self, keep_variance, whiten, random_state=None):
        self.keep_variance = keep_variance
        self.whiten = whiten

        self.random_state = random_state

    def fit(self, X, Y=None):
        import sklearn.decomposition

        n_components = float(self.keep_variance)
        self.whiten = check_for_bool(self.whiten)

        self.preprocessor = sklearn.decomposition.PCA(
            n_components=n_components, whiten=self.whiten, copy=True
        )
        self.preprocessor.fit(X)

        if not np.isfinite(self.preprocessor.components_).all():
            raise ValueError("PCA found non-finite components.")

        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "PCA",
            "name": "Principle Component Analysis",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            # TODO document that we have to be very careful
            "is_deterministic": False,
            "input": (DENSE, UNSIGNED_DATA),
            "output": (DENSE, UNSIGNED_DATA),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        keep_variance = UniformFloatHyperparameter(
            "keep_variance", 0.5, 0.9999, default_value=0.9999
        )
        whiten = CategoricalHyperparameter(
            "whiten", ["False", "True"], default_value="False"
        )
        cs = ConfigurationSpace()
        cs.add_hyperparameters([keep_variance, whiten])
        return cs



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import sklearn.compose
from ConfigSpace import Configuration
from ConfigSpace.configuration_space import ConfigurationSpace
from scipy import sparse
from sklearn.base import BaseEstimator

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES
from autosklearn.pipeline.base import (
    DATASET_PROPERTIES_TYPE,
    PIPELINE_DATA_DTYPE,
    BasePipeline,
)
from autosklearn.pipeline.components.base import (
    AutoSklearnChoice,
    AutoSklearnComponent,
    AutoSklearnPreprocessingAlgorithm,
)
from autosklearn.pipeline.components.data_preprocessing.feature_type_categorical import (  # noqa : E501
    CategoricalPreprocessingPipeline,
)
from autosklearn.pipeline.components.data_preprocessing.feature_type_numerical import (
    NumericalPreprocessingPipeline,
)
from autosklearn.pipeline.components.data_preprocessing.feature_type_text import (
    TextPreprocessingPipeline,
)
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class FeatTypeSplit(AutoSklearnPreprocessingAlgorithm):
    """
    This component is used to apply distinct transformations to categorical,
    numerical and text features of a dataset. It is built on top of sklearn's
    ColumnTransformer.
    """

    def __init__(
        self,
        config: Optional[Configuration] = None,
        pipeline: Optional[BasePipeline] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
        include: Optional[Dict[str, str]] = None,
        exclude: Optional[Dict[str, str]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        init_params: Optional[Dict[str, Any]] = None,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        force_sparse_output: bool = False,
        column_transformer: Optional[sklearn.compose.ColumnTransformer] = None,
    ):

        if pipeline is not None:
            raise ValueError("DataPreprocessor's argument 'pipeline' should be None")

        self.config = config
        self.pipeline = pipeline
        self.dataset_properties = dataset_properties
        self.include = include
        self.exclude = exclude
        self.random_state = random_state
        self.init_params = init_params
        self.feat_type = feat_type
        self.force_sparse_output = force_sparse_output

        # The pipeline that will be applied to the categorical features (i.e. columns)
        # of the dataset
        # Configuration of the data-preprocessor is different from the configuration of
        # the categorical pipeline. Hence, force to None
        # It is actually the call to set_hyperparameter who properly sets this argument
        # TODO: Extract the child configuration space from the FeatTypeSplit to the
        # pipeline if needed
        self.categ_ppl = CategoricalPreprocessingPipeline(
            feat_type=self.feat_type,
            config=None,
            steps=pipeline,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
        )
        # The pipeline that will be applied to the numerical features (i.e. columns)
        # of the dataset
        # Configuration of the data-preprocessor is different from the configuration of
        # the numerical pipeline. Hence, force to None
        # It is actually the call to set_hyperparameter who properly sets this argument
        # TODO: Extract the child configuration space from the FeatTypeSplit to the
        # pipeline if needed
        self.numer_ppl = NumericalPreprocessingPipeline(
            feat_type=self.feat_type,
            config=None,
            steps=pipeline,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
        )

        # The pipeline that will be applied to the text features (i.e. columns)
        # of the dataset
        # Configuration of the data-preprocessor is different from the configuration of
        # the numerical or categorical pipeline. Hence, force to None
        # It is actually the call to set_hyperparameter who properly sets this argument
        # TODO: Extract the child configuration space from the FeatTypeSplit to the
        # pipeline if needed
        self.txt_ppl = TextPreprocessingPipeline(
            feat_type=self.feat_type,
            config=None,
            steps=pipeline,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
        )

        if self.feat_type is None:
            self._transformers: List[Tuple[str, AutoSklearnComponent]] = [
                ("categorical_transformer", self.categ_ppl),
                ("numerical_transformer", self.numer_ppl),
                ("text_transformer", self.txt_ppl),
            ]
        else:
            self._transformers: List[Tuple[str, AutoSklearnComponent]] = []
            if "categorical" in self.feat_type.values():
                self._transformers.append(("categorical_transformer", self.categ_ppl))
            if "numerical" in self.feat_type.values():
                self._transformers.append(("numerical_transformer", self.numer_ppl))
            if "string" in self.feat_type.values():
                self._transformers.append(("text_transformer", self.txt_ppl))

        if self.config:
            self.set_hyperparameters(
                feat_type=self.feat_type,
                configuration=self.config,
                init_params=init_params,
            )

        self.column_transformer = column_transformer

    def fit(
        self, X: SUPPORTED_FEAT_TYPES, y: Optional[SUPPORTED_TARGET_TYPES] = None
    ) -> "FeatTypeSplit":

        n_feats = X.shape[1]
        if self.feat_type is not None:
            # Make sure that we are not missing any column!
            expected = set(self.feat_type.keys())
            if hasattr(X, "columns"):
                columns = set(X.columns)
            else:
                columns = set(range(n_feats))
            if expected != columns:
                raise ValueError(
                    f"Train data has columns={expected} yet the"
                    f" feat_types are feat={columns}"
                )
            sklearn_transf_spec = []

            categorical_features = [
                key
                for key, value in self.feat_type.items()
                if value.lower() == "categorical"
            ]
            if len(categorical_features) > 0:
                sklearn_transf_spec.append(
                    ("categorical_transformer", self.categ_ppl, categorical_features)
                )

            numerical_features = [
                key
                for key, value in self.feat_type.items()
                if value.lower() == "numerical"
            ]
            if len(numerical_features) > 0:
                sklearn_transf_spec.append(
                    ("numerical_transformer", self.numer_ppl, numerical_features)
                )

            text_features = [
                key
                for key, value in self.feat_type.items()
                if value.lower() == "string"
            ]
            if len(text_features) > 0:
                sklearn_transf_spec.append(
                    ("text_transformer", self.txt_ppl, text_features)
                )
        else:
            # self.feature_type == None assumes numerical case
            sklearn_transf_spec = [
                ("numerical_transformer", self.numer_ppl, [True] * n_feats)
            ]

        # And one last check in case feat type is None
        # And to make sure the final specification has all the columns
        # considered in the column transformer
        total_columns = sum(
            [len(features) for name, ppl, features in sklearn_transf_spec]
        )
        if total_columns != n_feats:
            raise ValueError(
                "Missing columns in the specification of the data validator"
                f" for train data={np.shape(X)} and spec={sklearn_transf_spec}"
            )

        self.sparse_ = sparse.issparse(X) or self.force_sparse_output
        self.column_transformer = sklearn.compose.ColumnTransformer(
            transformers=sklearn_transf_spec,
            sparse_threshold=float(self.sparse_),
        )
        self.column_transformer.fit(X, y)
        return self

    def transform(self, X: SUPPORTED_FEAT_TYPES) -> PIPELINE_DATA_DTYPE:
        if self.column_transformer is None:
            raise ValueError(
                "Cannot call transform on a Datapreprocessor that has not"
                "yet been fit. Please check the log files for errors "
                "while trying to fit the model."
            )
        return self.column_transformer.transform(X)

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "FeatTypeSplit",
            "name": "Feature Type Splitter",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            # TODO find out of this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    def set_hyperparameters(
        self,
        feat_type: FEAT_TYPE_TYPE,
        configuration: Configuration,
        init_params: Optional[Dict[str, Any]] = None,
    ) -> "FeatTypeSplit":
        if init_params is not None and "feat_type" in init_params.keys():
            self.feat_type = init_params["feat_type"]

        self.config = configuration

        for transf_name, transf_op in self._transformers:
            sub_configuration_space = transf_op.get_hyperparameter_search_space(
                dataset_properties=self.dataset_properties, feat_type=feat_type
            )
            sub_config_dict = {}
            for param in configuration:
                if param.startswith("%s:" % transf_name):
                    value = configuration[param]
                    new_name = param.replace("%s:" % transf_name, "", 1)
                    sub_config_dict[new_name] = value

            sub_configuration = Configuration(
                sub_configuration_space,
                values=sub_config_dict,
                allow_inactive_with_values=True,
            )

            sub_init_params_dict: Optional[Dict[str, Any]] = None
            if init_params is not None:
                sub_init_params_dict = {}
                for param in init_params:
                    if param.startswith("%s:" % transf_name):
                        value = init_params[param]
                        new_name = param.replace("%s:" % transf_name, "", 1)
                        sub_init_params_dict[new_name] = value

            if isinstance(
                transf_op, (AutoSklearnChoice, AutoSklearnComponent, BasePipeline)
            ):
                transf_op.set_hyperparameters(
                    feat_type=feat_type,
                    configuration=sub_configuration,
                    init_params=sub_init_params_dict,
                )
            else:
                raise NotImplementedError("Not supported yet!")

        return self

    def get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        self.dataset_properties = dataset_properties
        cs = ConfigurationSpace()
        cs = FeatTypeSplit._get_hyperparameter_search_space_recursevely(
            feat_type=feat_type,
            dataset_properties=dataset_properties,
            cs=cs,
            transformer=self._transformers,
        )
        return cs

    @staticmethod
    def _get_hyperparameter_search_space_recursevely(
        dataset_properties: DATASET_PROPERTIES_TYPE,
        cs: ConfigurationSpace,
        transformer: BaseEstimator,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
    ) -> ConfigurationSpace:
        for st_name, st_operation in transformer:
            if hasattr(st_operation, "get_hyperparameter_search_space"):
                cs.add_configuration_space(
                    st_name,
                    st_operation.get_hyperparameter_search_space(
                        dataset_properties=dataset_properties
                    ),
                )
            else:
                return FeatTypeSplit._get_hyperparameter_search_space_recursevely(
                    dataset_properties, cs, st_operation
                )
        return cs



import os
import sys

from autosklearn.pipeline.components.base import (
    AutoSklearnClassificationAlgorithm,
    find_components,
)

import unittest

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(this_dir)


class TestBase(unittest.TestCase):
    def test_find_components(self):
        c = find_components(
            "dummy_components",
            os.path.join(this_dir, "dummy_components"),
            AutoSklearnClassificationAlgorithm,
        )
        print("COMPONENTS: %s" % repr(c))
        self.assertEqual(len(c), 2)
        self.assertEqual(c["dummy_component_1"].__name__, "DummyComponent1")
        self.assertEqual(c["dummy_component_2"].__name__, "DummyComponent2")



from __future__ import annotations

from typing import Any, Callable

import os
import pickle
import shutil
import tempfile
from functools import partial
from pathlib import Path

from filelock import FileLock

from pytest import FixtureRequest
from pytest_cases import fixture

from test.conftest import AUTOSKLEARN_CACHE_NAME

LOCK_DIR = Path(tempfile.gettempdir())


class Cache:
    """Used for the below fixtures.

    Mainly used with cases so they don' need to be retrained at every invocation.
    The cache can be cleared using `pytest`'s built in mechanism:

        pytest --clear-cache

    To view cached items use:

        pytest --cache-show

    ..code:: python

        def case_fitted_model(cache, ...):
            key = "some key unique to this test"
            cache = cache(key)
            if "model" in cache:
                return cache.load("model")

            # ... fit model

            cache.save(model, "model")
            return model

    If multiple items are required, they can be access in much the same ways

    ..code:: python

        def case_requires_multiple_things(cache, ...):

            cache1 = cache("key1")
            cache2 = cache("key2")

    If multiple things need to be stored in one location, you can access a given path
    for a given named thing inside a cache.

    ..code:: python

        def case_fitted_model_and_populated_backend(cache, ...):
            cache = cache("some key")

    """

    def __init__(self, key: str, cache_dir: Path, verbose: int = 0):
        """
        Parameters
        ----------
        key : str
            The key of the item stored

        cache_dir : Path
            The dir where the cache should be located

        verbose : int = 0
            Whether to be verbose or not. Currently only has one level (> 0)
        """
        self.key = key
        self.dir = cache_dir / key
        self.verbose = verbose > 0
        self._lock: FileLock = None

    def items(self) -> list[Path]:
        """Get any paths associated to items in this dir"""
        return list(self.dir.iterdir())

    def __contains__(self, name: str) -> bool:
        return self.path(name).exists()

    def path(self, name: str) -> Path:
        """Path to an item for this cache"""
        return self.dir / name

    def load(self, name: str) -> Any:
        """Load an item from the cache with a given name"""
        if self.verbose:
            print(f"Loading cached item {self.path(name)}")

        with self.path(name).open("rb") as f:
            return pickle.load(f)

    def save(self, item: Any, name: str) -> None:
        """Dump an item to cache with a name"""
        if self.verbose:
            print(f"Saving cached item {self.path(name)}")

        with self.path(name).open("wb") as f:
            pickle.dump(item, f)

    def reset(self) -> None:
        """Delete this caches items"""
        shutil.rmtree(self.dir)
        self.dir.mkdir()

    def __enter__(self):
        if int(os.environ.get("PYTEST_XDIST_WORKER_COUNT", 1)) <= 1:
            return self
        else:
            path = LOCK_DIR / f"{self.key}.lock"
            self._lock = FileLock(path)
            self._lock.acquire(poll_interval=1.0)
            if self.verbose:
                print(f"locked cache {path}")

            return self

    def __exit__(self, *args, **kwargs):
        if self._lock is not None:
            self._lock.release()


@fixture
def make_cache(request: FixtureRequest) -> Callable[[str], Cache]:
    """Gives the access to a cache."""
    pytest_cache = request.config.cache
    assert pytest_cache is not None

    cache_dir = pytest_cache.mkdir(AUTOSKLEARN_CACHE_NAME)
    return partial(Cache, cache_dir=cache_dir)



from typing import Dict, Optional, Tuple, Union

import ConfigSpace.hyperparameters as CSH
import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from sklearn.decomposition import TruncatedSVD

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class TextFeatureReduction(AutoSklearnPreprocessingAlgorithm):
    """
    Reduces the features created by a bag of words encoding
    """

    def __init__(
        self,
        n_components: Optional[int] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        self.n_components = n_components
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "TextFeatureReduction":
        if X.shape[1] > self.n_components:
            self.preprocessor = TruncatedSVD(
                n_components=self.n_components, random_state=self.random_state
            )
            self.preprocessor.fit(X)
        elif X.shape[1] <= self.n_components and X.shape[1] != 1:
            self.preprocessor = TruncatedSVD(
                n_components=X.shape[1] - 1, random_state=self.random_state
            )
            self.preprocessor.fit(X)
        elif X.shape[1] == 1:
            self.preprocessor = "passthrough"
        else:
            raise ValueError(
                "The text embedding consists only of a single dimension.\n"
                "Are you sure that your text data is necessary?"
            )
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        elif self.preprocessor == "passthrough":
            return X
        else:
            return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "TextFeatureReduction",
            "name": "TextFeatureReduction",
            "handles_missing_values": True,
            "handles_nominal_values": True,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            CSH.UniformIntegerHyperparameter(
                "n_components", lower=1, upper=10000, default_value=100, log=True
            )
        )
        return cs



import numpy as np
import sklearn.naive_bayes
import sklearn.preprocessing

from autosklearn.pipeline.components.classification.multinomial_nb import MultinomialNB
from autosklearn.pipeline.util import get_dataset

from .test_base import BaseClassificationComponentTest


class MultinomialNBComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.97999999999999998
    res["iris_n_calls"] = None
    res["default_iris_iterative"] = 0.97999999999999998
    res["default_iris_proba"] = 0.5865733413579101
    res["default_iris_sparse"] = 0.54
    res["default_digits"] = 0.89496053430479661
    res["digits_n_calls"] = None
    res["default_digits_iterative"] = 0.89496053430479661
    res["default_digits_binary"] = 0.98967820279295693
    res["default_digits_multilabel"] = 0.70484946987667163
    res["default_digits_multilabel_proba"] = 0.80324074074074081

    sk_mod = sklearn.naive_bayes.MultinomialNB
    module = MultinomialNB

    def test_default_configuration_negative_values(self):
        # Custon preprocessing test to check if clipping to zero works
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="digits")
        ss = sklearn.preprocessing.StandardScaler()
        X_train = ss.fit_transform(X_train)
        configuration_space = MultinomialNB.get_hyperparameter_search_space()
        default = configuration_space.get_default_configuration()

        cls = MultinomialNB(
            random_state=1,
            **{
                hp_name: default[hp_name]
                for hp_name in default
                if default[hp_name] is not None
            },
        )

        cls = cls.fit(X_train, Y_train)
        prediction = cls.predict(X_test)
        self.assertAlmostEqual(np.nanmean(prediction == Y_test), 0.88888888888888884)



import os
import sys

from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm

# Add the parent directory to the path to import the parent component as
# dummy_components.dummy_component_2.DummyComponent1
this_directory = os.path.dirname(os.path.abspath(__file__))
parent_directory = os.path.abspath(os.path.join(this_directory, ".."))
sys.path.append(parent_directory)


class DummyComponent2(AutoSklearnClassificationAlgorithm):
    pass



import copy
import itertools
import resource
import tempfile

import numpy as np
import sklearn.datasets
import sklearn.decomposition
import sklearn.ensemble
import sklearn.svm
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter
from joblib import Memory
from sklearn.base import clone
from sklearn.utils.validation import check_is_fitted

import autosklearn.pipeline.components.feature_preprocessing as preprocessing_components
import autosklearn.pipeline.components.regression as regression_components
from autosklearn.pipeline.components.base import (
    AutoSklearnChoice,
    AutoSklearnComponent,
    AutoSklearnPreprocessingAlgorithm,
    AutoSklearnRegressionAlgorithm,
)
from autosklearn.pipeline.constants import (
    DENSE,
    PREDICTIONS,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)
from autosklearn.pipeline.regression import SimpleRegressionPipeline
from autosklearn.pipeline.util import get_dataset

import unittest
import unittest.mock

from test.test_pipeline.ignored_warnings import ignore_warnings, regressor_warnings


class SimpleRegressionPipelineTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_io_dict(self):
        regressors = regression_components._regressors
        for r in regressors:
            if regressors[r] == regression_components.RegressorChoice:
                continue
            props = regressors[r].get_properties()
            self.assertIn("input", props)
            self.assertIn("output", props)
            inp = props["input"]
            output = props["output"]

            self.assertIsInstance(inp, tuple)
            self.assertIsInstance(output, tuple)
            for i in inp:
                self.assertIn(i, (SPARSE, DENSE, SIGNED_DATA, UNSIGNED_DATA))
            self.assertEqual(output, (PREDICTIONS,))
            self.assertIn("handles_regression", props)
            self.assertTrue(props["handles_regression"])
            self.assertIn("handles_classification", props)
            self.assertIn("handles_multiclass", props)
            self.assertIn("handles_multilabel", props)
            self.assertIn("handles_multioutput", props)
            self.assertFalse(props["handles_classification"])
            self.assertFalse(props["handles_multiclass"])
            self.assertFalse(props["handles_multilabel"])

    def test_find_regressors(self):
        regressors = regression_components._regressors
        self.assertGreaterEqual(len(regressors), 1)
        for key in regressors:
            if hasattr(regressors[key], "get_components"):
                continue
            self.assertIn(AutoSklearnRegressionAlgorithm, regressors[key].__bases__)

    def test_find_preprocessors(self):
        preprocessors = preprocessing_components._preprocessors
        self.assertGreaterEqual(len(preprocessors), 1)
        for key in preprocessors:
            if hasattr(preprocessors[key], "get_components"):
                continue
            self.assertIn(
                AutoSklearnPreprocessingAlgorithm, preprocessors[key].__bases__
            )

    def test_configurations(self):
        cs = SimpleRegressionPipeline().get_hyperparameter_search_space()

        self._test_configurations(cs)

    def test_configurations_signed_data(self):
        dataset_properties = {"signed": True}
        cs = SimpleRegressionPipeline(
            dataset_properties=dataset_properties
        ).get_hyperparameter_search_space()

        self._test_configurations(
            configurations_space=cs, dataset_properties=dataset_properties
        )

    def test_configurations_sparse(self):
        dataset_properties = {"sparse": True}
        cs = SimpleRegressionPipeline(
            dataset_properties=dataset_properties
        ).get_hyperparameter_search_space()

        self._test_configurations(
            cs, make_sparse=True, dataset_properties=dataset_properties
        )

    def test_multioutput(self):
        cache = Memory(location=tempfile.gettempdir())
        cached_func = cache.cache(sklearn.datasets.make_regression)
        X, Y = cached_func(
            n_samples=250,
            n_features=20,
            n_informative=9,
            n_targets=4,
            bias=0.5,
            effective_rank=10,
            tail_strength=0.4,
            noise=0.3,
            shuffle=True,
            coef=False,
            random_state=1,
        )
        X_train = X[:200, :]
        Y_train = Y[:200, :]
        X_test = X[200:, :]
        Y_test = Y[200:, :]

        data = {
            "X_train": X_train,
            "Y_train": Y_train,
            "X_test": X_test,
            "Y_test": Y_test,
        }

        dataset_properties = {"multioutput": True}
        pipeline = SimpleRegressionPipeline(dataset_properties=dataset_properties)
        cs = pipeline.get_hyperparameter_search_space()

        self._test_configurations(cs, data=data, dataset_properties=dataset_properties)

    def _test_configurations(
        self,
        configurations_space,
        make_sparse=False,
        data=None,
        dataset_properties=None,
    ):
        # Use a limit of ~4GiB
        limit = 3072 * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (limit, limit))

        configurations_space.seed(1)

        for i in range(10):
            config = configurations_space.sample_configuration()
            config._populate_values()

            # Restrict configurations which could take too long on travis-ci
            restrictions = {
                "regressor:adaboost:n_estimators": 50,
                "regressor:adaboost:max_depth": 1,
                "feature_preprocessor:kernel_pca:n_components": 10,
                "feature_preprocessor:kitchen_sinks:n_components": 50,
                "regressor:libsvm_svc:degree": 2,
                "regressor:libsvm_svr:degree": 2,
                "regressor:libsvm_svr:C": 1.0,
                "feature_preprocessor:truncatedSVD:target_dim": 10,
                "feature_preprocessor:polynomial:degree": 2,
                "regressor:lda:n_components": 10,
            }

            for restrict_parameter in restrictions:
                restrict_to = restrictions[restrict_parameter]
                if (
                    restrict_parameter in config
                    and config[restrict_parameter] is not None
                ):
                    config._values[restrict_parameter] = restrict_to

            if data is None:
                X_train, Y_train, X_test, Y_test = get_dataset(
                    dataset="boston", make_sparse=make_sparse, add_NaNs=True
                )
            else:
                X_train = data["X_train"].copy()
                Y_train = data["Y_train"].copy()
                X_test = data["X_test"].copy()
                data["Y_test"].copy()

            cls = SimpleRegressionPipeline(
                random_state=1, dataset_properties=dataset_properties
            )
            cls.set_hyperparameters(config)

            # First make sure that for this configuration, setting the parameters
            # does not mistakenly set the estimator as fitted
            for name, step in cls.named_steps.items():
                with self.assertRaisesRegex(
                    sklearn.exceptions.NotFittedError, "instance is not fitted yet"
                ):
                    check_is_fitted(step)

            try:
                with ignore_warnings(regressor_warnings):
                    cls.fit(X_train, Y_train)

                # After fit, all components should be tagged as fitted
                # by sklearn. Check is fitted raises an exception if that
                # is not the case
                try:
                    for name, step in cls.named_steps.items():
                        check_is_fitted(step)
                except sklearn.exceptions.NotFittedError:
                    self.fail(
                        "config={} raised NotFittedError unexpectedly!".format(config)
                    )

                cls.predict(X_test)
            except MemoryError:
                continue
            except np.linalg.LinAlgError:
                continue
            except ValueError as e:
                if "Floating-point under-/overflow occurred at epoch" in e.args[0]:
                    continue
                elif "removed all features" in e.args[0]:
                    continue
                elif "all features are discarded" in e.args[0]:
                    continue
                elif "Numerical problems in QDA" in e.args[0]:
                    continue
                elif "Bug in scikit-learn" in e.args[0]:
                    continue
                elif (
                    "The condensed distance matrix must contain only finite "
                    "values." in e.args[0]
                ):
                    continue
                elif (
                    "zero-size array to reduction operation maximum which has no "
                    "identity" in e.args[0]
                ):
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

            except RuntimeWarning as e:
                if "invalid value encountered in sqrt" in e.args[0]:
                    continue
                elif "divide by zero encountered in" in e.args[0]:
                    continue
                elif "invalid value encountered in divide" in e.args[0]:
                    continue
                elif "invalid value encountered in true_divide" in e.args[0]:
                    continue
                elif "invalid value encountered in multiply" in e.args[0]:
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

            except UserWarning as e:
                if "FastICA did not converge" in e.args[0]:
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

            except Exception as e:
                if (
                    "Multiple input features cannot have the same target value"
                    in e.args[0]
                ):
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

    def test_default_configuration(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset="diabetes")
            auto = SimpleRegressionPipeline(random_state=1)
            auto = auto.fit(X_train, Y_train)
            predictions = auto.predict(copy.deepcopy(X_test))
            # The lower the worse
            r2_score = sklearn.metrics.r2_score(Y_test, predictions)
            self.assertAlmostEqual(0.3458397471855429, r2_score, places=2)
            model_score = auto.score(copy.deepcopy(X_test), Y_test)
            self.assertAlmostEqual(model_score, r2_score, places=5)

    def test_default_configuration_iterative_fit(self):
        regressor = SimpleRegressionPipeline(
            random_state=1,
            include={
                "regressor": ["random_forest"],
                "feature_preprocessor": ["no_preprocessing"],
            },
        )
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="boston")
        regressor.fit_transformer(X_train, Y_train)
        for i in range(1, 11):
            regressor.iterative_fit(X_train, Y_train)
            self.assertEqual(regressor.steps[-1][-1].choice.estimator.n_estimators, i)

    def test_repr(self):
        representation = repr(SimpleRegressionPipeline())
        cls = eval(representation)
        self.assertIsInstance(cls, SimpleRegressionPipeline)

    def test_get_hyperparameter_search_space(self):
        cs = SimpleRegressionPipeline().get_hyperparameter_search_space()
        self.assertIsInstance(cs, ConfigurationSpace)
        conditions = cs.get_conditions()
        hyperparameters = cs.get_hyperparameters()
        forbiddens = cs.get_forbiddens()
        self.assertEqual(167, len(hyperparameters))
        self.assertEqual(len(hyperparameters) - 3, len(conditions))
        self.assertEqual(len(forbiddens), 35)

    def test_get_hyperparameter_search_space_include_exclude_models(self):
        regressor = SimpleRegressionPipeline(include={"regressor": ["random_forest"]})
        cs = regressor.get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter("regressor:__choice__"),
            CategoricalHyperparameter("regressor:__choice__", ["random_forest"]),
        )

        # TODO add this test when more than one regressor is present
        regressor = SimpleRegressionPipeline(exclude={"regressor": ["random_forest"]})
        cs = regressor.get_hyperparameter_search_space()
        self.assertNotIn("random_forest", str(cs))

        regressor = SimpleRegressionPipeline(include={"feature_preprocessor": ["pca"]})
        cs = regressor.get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter("feature_preprocessor:__choice__"),
            CategoricalHyperparameter("feature_preprocessor:__choice__", ["pca"]),
        )

        regressor = SimpleRegressionPipeline(
            exclude={"feature_preprocessor": ["no_preprocessing"]}
        )
        cs = regressor.get_hyperparameter_search_space()
        self.assertNotIn("no_preprocessing", str(cs))

    def test_get_hyperparameter_search_space_preprocessor_contradicts_default(
        self,
    ):
        regressor = SimpleRegressionPipeline(
            include={"feature_preprocessor": ["densifier"]},
            dataset_properties={"sparse": True},
        )
        cs = regressor.get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter("regressor:__choice__").default_value,
            "gradient_boosting",
        )

        regressor = SimpleRegressionPipeline(
            include={"feature_preprocessor": ["nystroem_sampler"]}
        )
        cs = regressor.get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter("regressor:__choice__").default_value, "sgd"
        )

    def test_get_hyperparameter_search_space_only_forbidden_combinations(self):
        self.assertRaisesRegex(
            ValueError,
            "Cannot find a legal default configuration.",
            SimpleRegressionPipeline,
            include={
                "regressor": ["random_forest"],
                "feature_preprocessor": ["kitchen_sinks"],
            },
        )

        # It must also be catched that no classifiers which can handle sparse
        # data are located behind the densifier
        self.assertRaisesRegex(
            ValueError,
            "Cannot find a legal default configuration",
            SimpleRegressionPipeline,
            include={
                "regressor": ["extra_trees"],
                "feature_preprocessor": ["densifier"],
            },
            dataset_properties={"sparse": True},
        )

    @unittest.skip(
        "test_get_hyperparameter_search_space_dataset_properties"
        + " Not yet Implemented"
    )
    def test_get_hyperparameter_search_space_dataset_properties(self):
        # TODO: We do not have any dataset properties for regression, so this
        # test is somewhat stupid
        pass
        """
        full_cs = SimpleRegressionPipeline.get_hyperparameter_search_space()
        cs_mc = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertEqual(full_cs, cs_mc)

        cs_ml = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertNotIn('k_nearest_neighbors', str(cs_ml))
        self.assertNotIn('liblinear', str(cs_ml))
        self.assertNotIn('libsvm_svc', str(cs_ml))
        self.assertNotIn('sgd', str(cs_ml))

        cs_sp = SimpleRegressionPipeline.get_hyperparameter_search_space(
            sparse=True)
        self.assertNotIn('extra_trees', str(cs_sp))
        self.assertNotIn('gradient_boosting', str(cs_sp))
        self.assertNotIn('random_forest', str(cs_sp))

        cs_mc_ml = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertEqual(cs_ml, cs_mc_ml)

        self.assertRaisesRegex(ValueError,
                                "No regressor to build a configuration space "
                                "for...", SimpleRegressionPipeline.
                                get_hyperparameter_search_space,
                                multiclass=True, multilabel=True, sparse=True)
    """

    def test_predict_batched(self):
        include = {"regressor": ["decision_tree"]}
        cs = SimpleRegressionPipeline(include=include).get_hyperparameter_search_space()
        default = cs.get_default_configuration()
        regressor = SimpleRegressionPipeline(
            config=default, random_state=1, include=include
        )

        X_train, Y_train, X_test, Y_test = get_dataset(dataset="boston")
        regressor.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = regressor.predict(X_test_)
        mock_predict = unittest.mock.Mock(wraps=regressor.steps[-1][-1].predict)
        regressor.steps[-1][-1].predict = mock_predict
        prediction = regressor.predict(X_test, batch_size=20)
        self.assertEqual((356,), prediction.shape)
        self.assertEqual(18, mock_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    def test_predict_batched_sparse(self):
        dataset_properties = {"sparse": True}
        include = {"regressor": ["decision_tree"]}

        cs = SimpleRegressionPipeline(
            dataset_properties=dataset_properties, include=include
        ).get_hyperparameter_search_space()

        default = cs.get_default_configuration()
        regressor = SimpleRegressionPipeline(
            config=default,
            random_state=1,
            dataset_properties=dataset_properties,
            include=include,
        )

        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset="boston", make_sparse=True
        )
        regressor.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = regressor.predict(X_test_)
        mock_predict = unittest.mock.Mock(wraps=regressor.steps[-1][-1].predict)
        regressor.steps[-1][-1].predict = mock_predict
        prediction = regressor.predict(X_test, batch_size=20)
        self.assertEqual((356,), prediction.shape)
        self.assertEqual(18, mock_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    @unittest.skip("test_check_random_state Not yet Implemented")
    def test_check_random_state(self):
        raise NotImplementedError()

    @unittest.skip("test_validate_input_X Not yet Implemented")
    def test_validate_input_X(self):
        raise NotImplementedError()

    @unittest.skip("test_validate_input_Y Not yet Implemented")
    def test_validate_input_Y(self):
        raise NotImplementedError()

    def test_pipeline_clonability(self):
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="boston")
        auto = SimpleRegressionPipeline(random_state=1)
        auto = auto.fit(X_train, Y_train)
        auto_clone = clone(auto)
        auto_clone_params = auto_clone.get_params()

        # Make sure all keys are copied properly
        for k, v in auto.get_params().items():
            self.assertIn(k, auto_clone_params)

        # Make sure the params getter of estimator are honored
        klass = auto.__class__
        new_object_params = auto.get_params(deep=False)
        for name, param in new_object_params.items():
            new_object_params[name] = clone(param, safe=False)
        new_object = klass(**new_object_params)
        params_set = new_object.get_params(deep=False)

        for name in new_object_params:
            param1 = new_object_params[name]
            param2 = params_set[name]
            self.assertEqual(param1, param2)

    def test_set_params(self):
        pass

    def test_get_params(self):
        pass

    def _test_set_hyperparameter_choice(
        self, expected_key, implementation, config_dict
    ):
        """
        Given a configuration in config, this procedure makes sure that
        the given implementation, which should be a Choice component, honors
        the type of the object, and any hyperparameter associated to it
        """
        keys_checked = [expected_key]
        implementation_type = config_dict[expected_key]
        expected_type = implementation.get_components()[implementation_type]
        self.assertIsInstance(implementation.choice, expected_type)

        # Are there further hyperparams?
        # A choice component might have attribute requirements that we need to check
        expected_sub_key = (
            expected_key.replace(":__choice__", ":") + implementation_type
        )
        expected_attributes = {}
        if "data_preprocessor:__choice__" in expected_key:
            # We have to check both the numerical and categorical
            to_check = {
                "numerical_transformer": implementation.choice.numer_ppl.named_steps,
                "categorical_transformer": implementation.choice.categ_ppl.named_steps,
                "text_transformer": implementation.choice.txt_ppl.named_steps,
            }

            for data_type, pipeline in to_check.items():
                for sub_name, sub_step in pipeline.items():
                    # If it is a Choice, make sure it is the correct one!
                    if isinstance(sub_step, AutoSklearnChoice):
                        key = "data_preprocessor:feature_type:{}:{}:__choice__".format(
                            data_type, sub_name
                        )
                        keys_checked.extend(
                            self._test_set_hyperparameter_choice(
                                key, sub_step, config_dict
                            )
                        )
                    # If it is a component, make sure it has the correct hyperparams
                    elif isinstance(sub_step, AutoSklearnComponent):
                        keys_checked.extend(
                            self._test_set_hyperparameter_component(
                                "data_preprocessor:feature_type:{}:{}".format(
                                    data_type, sub_name
                                ),
                                sub_step,
                                config_dict,
                            )
                        )
                    else:
                        raise ValueError("New type of pipeline component!")
            return keys_checked
        else:
            for key, value in config_dict.items():
                if key != expected_key and expected_sub_key in key:
                    expected_attributes[key.split(":")[-1]] = value
                    keys_checked.append(key)
        if expected_attributes:
            attributes = vars(implementation.choice)
            # Cannot check the whole dictionary, just names, as some
            # classes map the text hyperparameter directly to a function!
            for expected_attribute in expected_attributes.keys():
                self.assertIn(expected_attribute, attributes.keys())
        return keys_checked

    def _test_set_hyperparameter_component(
        self, expected_key, implementation, config_dict
    ):
        """
        Given a configuration in config, this procedure makes sure that
        the given implementation, which should be a autosklearn component, honors
        is created with the desired hyperparameters stated in config_dict
        """
        keys_checked = []
        attributes = vars(implementation)
        expected_attributes = {}
        for key, value in config_dict.items():
            if expected_key in key:
                keys_checked.append(key)
                key = key.replace(expected_key + ":", "")
                if ":" in key:
                    raise ValueError(
                        "This utility should only be called with a "
                        "matching string that produces leaf configurations, "
                        "that is no further colons are expected, yet key={}"
                        "".format(key)
                    )
                expected_attributes[key] = value
        # Cannot check the whole dictionary, just names, as some
        # classes map the text hyperparameter directly to a function!
        for expected_attribute in expected_attributes.keys():
            self.assertIn(expected_attribute, attributes.keys())
        return keys_checked

    def test_set_hyperparameters_honors_configuration(self):
        """Makes sure that a given configuration is honored in practice.

        This method tests that the set hyperparameters actually create objects
        that comply with the given configuration. It iterates trough the pipeline to
        make sure we did not miss a step, but also checks at the end that every
        configuration from Config was checked

        Also considers random_state and ensures pipeline steps correctly recieve
        the right random_state
        """

        all_combinations = list(itertools.product([True, False], repeat=4))
        for (
            sparse,
            multilabel,
            signed,
            multiclass,
        ) in all_combinations:
            dataset_properties = {
                "sparse": sparse,
                "multilabel": multilabel,
                "multiclass": multiclass,
                "signed": signed,
            }
            random_state = 1
            auto = SimpleRegressionPipeline(
                random_state=random_state,
                dataset_properties=dataset_properties,
            )
            cs = auto.get_hyperparameter_search_space()
            config = cs.sample_configuration()

            # Set hyperparameters takes a given config and translate
            # a config to an actual implementation
            auto.set_hyperparameters(config)
            config_dict = config.get_dictionary()

            # keys to check is our mechanism to ensure that every
            # every config key is checked
            keys_checked = []

            for name, step in auto.named_steps.items():
                if name == "data_preprocessor":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "data_preprocessor:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                elif name == "feature_preprocessor":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "feature_preprocessor:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                elif name == "regressor":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "regressor:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                else:
                    raise ValueError(
                        "Found another type of step! Need to update this check"
                        " {}. ".format(name)
                    )

            # Make sure we checked the whole configuration
            self.assertSetEqual(set(config_dict.keys()), set(keys_checked))



# -*- encoding: utf-8 -*-
from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import functools
import json
import logging
import math
import multiprocessing
import time
import traceback
from queue import Empty

import numpy as np
import pynisher
from ConfigSpace import Configuration
from sklearn.model_selection._split import (
    BaseCrossValidator,
    BaseShuffleSplit,
    _RepeatedSplits,
)
from smac.runhistory.runhistory import RunInfo, RunValue
from smac.stats.stats import Stats
from smac.tae import StatusType, TAEAbortException
from smac.tae.execute_func import AbstractTAFunc

import autosklearn.evaluation.test_evaluator
import autosklearn.evaluation.train_evaluator
import autosklearn.evaluation.util
import autosklearn.pipeline.components
from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.evaluation.train_evaluator import TYPE_ADDITIONAL_INFO
from autosklearn.metrics import Scorer
from autosklearn.util.logging_ import PickableLoggerAdapter, get_named_client_logger
from autosklearn.util.parallel import preload_modules


def fit_predict_try_except_decorator(
    ta: Callable, queue: multiprocessing.Queue, cost_for_crash: float, **kwargs: Any
) -> None:

    try:
        return ta(queue=queue, **kwargs)
    except Exception as e:
        if isinstance(e, (MemoryError, pynisher.TimeoutException)):
            # Re-raise the memory error to let the pynisher handle that correctly
            raise e

        exception_traceback = traceback.format_exc()
        error_message = repr(e)

        # Printing stuff to stdout just in case the queue doesn't work,
        # which happened with the
        # following traceback:
        #     File "auto-sklearn/autosklearn/evaluation/__init__.py", line 29, in fit_predict_try_except_decorator  # noqa E501
        #     return ta(queue=queue, **kwargs)
        #     File "auto-sklearn/autosklearn/evaluation/train_evaluator.py", line 1067, in eval_holdout  # noqa E501
        #     evaluator.fit_predict_and_loss(iterative=iterative)
        #     File "auto-sklearn/autosklearn/evaluation/train_evaluator.py", line 616, in fit_predict_and_loss,  # noqa E501
        #     status=status
        #     File "auto-sklearn/autosklearn/evaluation/abstract_evaluator.py", line 320, in finish_up  # noqa E501
        #     self.queue.put(return_value_dict)
        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/multiprocessing/queues.py", line 87, in put  # noqa E501
        #     self._start_thread()
        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/multiprocessing/queues.py", line 170, in _start_thread  # noqa E501
        #     self._thread.start()
        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/threading.py", line 847, in start  # noqa E501
        #     RuntimeError: can't start new thread
        print(
            "Exception handling in `fit_predict_try_except_decorator`: "
            "traceback: %s \nerror message: %s" % (exception_traceback, error_message)
        )

        queue.put(
            {
                "loss": cost_for_crash,
                "additional_run_info": {
                    "traceback": exception_traceback,
                    "error": error_message,
                },
                "status": StatusType.CRASHED,
                "final_queue_element": True,
            },
            block=True,
        )
        queue.close()


def get_cost_of_crash(metrics: Sequence[Scorer]) -> List[float] | float:
    """Return the cost of crash.

    Return value can be either a list (multi-objective optimization) or a
    raw float (single objective) because SMAC assumes different types in the
    two different cases.
    """
    costs = []
    for metric in metrics:
        if not isinstance(metric, Scorer):
            raise ValueError("The metric {metric} must be an instance of Scorer")

        # Autosklearn optimizes the err. This function translates
        # worst_possible_result to be a minimization problem.
        # For metrics like accuracy that are bounded to [0,1]
        # metric.optimum==1 is the worst cost.
        # A simple guide is to use greater_is_better embedded as sign
        if metric._sign < 0:
            worst_possible_result = metric._worst_possible_result
        else:
            worst_possible_result = metric._optimum - metric._worst_possible_result
        costs.append(worst_possible_result)

    return costs if len(costs) > 1 else costs[0]


def _encode_exit_status(
    exit_status: Union[str, int, Type[BaseException]]
) -> Union[str, int]:
    try:
        # If it can be dumped, then it is int
        exit_status = cast(int, exit_status)
        json.dumps(exit_status)
        return exit_status
    except (TypeError, OverflowError):
        return str(exit_status)


# TODO potentially log all inputs to this class to pickle them in order to do
# easier debugging of potential crashes
class ExecuteTaFuncWithQueue(AbstractTAFunc):
    def __init__(
        self,
        backend: Backend,
        autosklearn_seed: int,
        resampling_strategy: Union[
            str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
        ],
        metrics: Sequence[Scorer],
        cost_for_crash: float,
        abort_on_first_run_crash: bool,
        port: int,
        pynisher_context: str,
        multi_objectives: List[str],
        initial_num_run: int = 1,
        stats: Optional[Stats] = None,
        run_obj: str = "quality",
        par_factor: int = 1,
        scoring_functions: Optional[List[Scorer]] = None,
        output_y_hat_optimization: bool = True,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        memory_limit: Optional[int] = None,
        disable_file_output: bool = False,
        init_params: Optional[Dict[str, Any]] = None,
        budget_type: Optional[str] = None,
        ta: Optional[Callable] = None,  # Required by SMAC's parent class
        **resampling_strategy_args: Any,
    ):
        if resampling_strategy == "holdout":
            eval_function = autosklearn.evaluation.train_evaluator.eval_holdout
        elif resampling_strategy == "holdout-iterative-fit":
            eval_function = (
                autosklearn.evaluation.train_evaluator.eval_iterative_holdout
            )
        elif resampling_strategy == "cv-iterative-fit":
            eval_function = autosklearn.evaluation.train_evaluator.eval_iterative_cv
        elif resampling_strategy == "cv" or isinstance(
            resampling_strategy, (BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit)
        ):
            eval_function = autosklearn.evaluation.train_evaluator.eval_cv
        elif resampling_strategy == "partial-cv":
            eval_function = autosklearn.evaluation.train_evaluator.eval_partial_cv
        elif resampling_strategy == "partial-cv-iterative-fit":
            eval_function = (
                autosklearn.evaluation.train_evaluator.eval_partial_cv_iterative
            )
        elif resampling_strategy == "test":
            eval_function = autosklearn.evaluation.test_evaluator.eval_t
            output_y_hat_optimization = False
        else:
            raise ValueError("Unknown resampling strategy %s" % resampling_strategy)

        self.worst_possible_result = cost_for_crash

        eval_function = functools.partial(
            fit_predict_try_except_decorator,
            ta=eval_function,
            cost_for_crash=self.worst_possible_result,
        )

        super().__init__(
            ta=eval_function,
            stats=stats,
            run_obj=run_obj,
            par_factor=par_factor,
            cost_for_crash=self.worst_possible_result,
            abort_on_first_run_crash=abort_on_first_run_crash,
            multi_objectives=multi_objectives,
        )

        self.backend = backend
        self.autosklearn_seed = autosklearn_seed
        self.resampling_strategy = resampling_strategy
        self.initial_num_run = initial_num_run
        self.metrics = metrics
        self.resampling_strategy = resampling_strategy
        self.resampling_strategy_args = resampling_strategy_args
        self.scoring_functions = scoring_functions
        # TODO deactivate output_y_hat_optimization and let respective evaluator decide
        self.output_y_hat_optimization = output_y_hat_optimization
        self.include = include
        self.exclude = exclude
        self.disable_file_output = disable_file_output
        self.init_params = init_params
        self.budget_type = budget_type

        if memory_limit is not None:
            memory_limit = int(math.ceil(memory_limit))
        self.memory_limit = memory_limit

        dm = self.backend.load_datamanager()
        self._get_test_loss = "X_test" in dm.data and "Y_test" in dm.data

        self.port = port
        self.pynisher_context = pynisher_context
        if self.port is None:
            self.logger: Union[
                logging.Logger, PickableLoggerAdapter
            ] = logging.getLogger("TAE")
        else:
            self.logger = get_named_client_logger(
                name="TAE",
                port=self.port,
            )

    def run_wrapper(
        self,
        run_info: RunInfo,
    ) -> Tuple[RunInfo, RunValue]:
        """
        wrapper function for ExecuteTARun.run_wrapper() to cap the target algorithm
        runtime if it would run over the total allowed runtime.

        Parameters
        ----------
        run_info : RunInfo
            Object that contains enough information to execute a configuration run in
            isolation.
        Returns
        -------
        RunInfo:
            an object containing the configuration launched
        RunValue:
            Contains information about the status/performance of config
        """
        if self.budget_type is None:
            if run_info.budget != 0:
                raise ValueError(
                    "If budget_type is None, budget must be.0, but is %f"
                    % run_info.budget
                )
        else:
            if run_info.budget == 0:
                run_info = run_info._replace(budget=100)
            elif run_info.budget <= 0 or run_info.budget > 100:
                raise ValueError(
                    "Illegal value for budget, must be >0 and <=100, but is %f"
                    % run_info.budget
                )
            if self.budget_type not in ("subsample", "iterations", "mixed"):
                raise ValueError(
                    "Illegal value for budget type, must be one of "
                    "('subsample', 'iterations', 'mixed'), but is : %s"
                    % self.budget_type
                )

        remaining_time = self.stats.get_remaing_time_budget()

        if remaining_time - 5 < run_info.cutoff:
            run_info = run_info._replace(cutoff=int(remaining_time - 5))

        config_id = (
            run_info.config
            if isinstance(run_info.config, int)
            else run_info.config.config_id
        )

        if run_info.cutoff < 1.0:
            self.logger.info(
                "Not starting configuration %d because time is up" % config_id
            )
            return run_info, RunValue(
                status=StatusType.STOP,
                cost=self.worst_possible_result,
                time=0.0,
                additional_info={},
                starttime=time.time(),
                endtime=time.time(),
            )
        elif run_info.cutoff != int(np.ceil(run_info.cutoff)) and not isinstance(
            run_info.cutoff, int
        ):
            run_info = run_info._replace(cutoff=int(np.ceil(run_info.cutoff)))

        self.logger.info("Starting to evaluate configuration %d" % config_id)
        return super().run_wrapper(run_info=run_info)

    def run(
        self,
        config: Configuration,
        instance: Optional[str] = None,
        cutoff: Optional[float] = None,
        seed: int = 12345,
        budget: float = 0.0,
        instance_specific: Optional[str] = None,
    ) -> Tuple[
        StatusType,
        float | list[float],
        float,
        Dict[str, Union[int, float, str, Dict, List, Tuple]],
    ]:

        # Additional information of each of the tae executions
        # Defined upfront for mypy
        additional_run_info: TYPE_ADDITIONAL_INFO = {}

        context = multiprocessing.get_context(self.pynisher_context)
        preload_modules(context)
        queue = context.Queue()

        if not (instance_specific is None or instance_specific == "0"):
            raise ValueError(instance_specific)

        init_params = {"instance": instance}
        if self.init_params is not None:
            init_params.update(self.init_params)

        if self.port is None:
            logger: Union[logging.Logger, PickableLoggerAdapter] = logging.getLogger(
                "pynisher"
            )
        else:
            logger = get_named_client_logger(
                name="pynisher",
                port=self.port,
            )
        arguments = dict(
            logger=logger,
            wall_time_in_s=cutoff,
            mem_in_mb=self.memory_limit,
            capture_output=True,
            context=context,
        )

        if isinstance(config, int):
            num_run = self.initial_num_run
        else:
            num_run = config.config_id + self.initial_num_run

        obj_kwargs = dict(
            queue=queue,
            config=config,
            backend=self.backend,
            port=self.port,
            metrics=self.metrics,
            seed=self.autosklearn_seed,
            num_run=num_run,
            scoring_functions=self.scoring_functions,
            output_y_hat_optimization=self.output_y_hat_optimization,
            include=self.include,
            exclude=self.exclude,
            disable_file_output=self.disable_file_output,
            instance=instance,
            init_params=init_params,
            budget=budget,
            budget_type=self.budget_type,
            additional_components=autosklearn.pipeline.components.base._addons,
        )

        if self.resampling_strategy != "test":
            obj_kwargs["resampling_strategy"] = self.resampling_strategy
            obj_kwargs["resampling_strategy_args"] = self.resampling_strategy_args

        try:
            obj = pynisher.enforce_limits(**arguments)(self.ta)
            obj(**obj_kwargs)
        except Exception as e:
            exception_traceback = traceback.format_exc()
            error_message = repr(e)
            additional_run_info.update(
                {"traceback": exception_traceback, "error": error_message}
            )
            return (
                StatusType.CRASHED,
                self.worst_possible_result,
                0.0,
                additional_run_info,
            )

        if obj.exit_status in (
            pynisher.TimeoutException,
            pynisher.MemorylimitException,
        ):
            # Even if the pynisher thinks that a timeout or memout occured,
            # it can be that the target algorithm wrote something into the queue
            #  - then we treat it as a succesful run
            try:
                info = autosklearn.evaluation.util.read_queue(queue)
                result = info[-1]["loss"]
                status = info[-1]["status"]
                additional_run_info = info[-1]["additional_run_info"]

                if obj.stdout:
                    additional_run_info["subprocess_stdout"] = obj.stdout
                if obj.stderr:
                    additional_run_info["subprocess_stderr"] = obj.stderr

                if obj.exit_status is pynisher.TimeoutException:
                    additional_run_info["info"] = "Run stopped because of timeout."
                elif obj.exit_status is pynisher.MemorylimitException:
                    additional_run_info["info"] = "Run stopped because of memout."

                if status in [StatusType.SUCCESS, StatusType.DONOTADVANCE]:
                    cost = result
                else:
                    cost = self.worst_possible_result

            except Empty:
                info = None
                if obj.exit_status is pynisher.TimeoutException:
                    status = StatusType.TIMEOUT
                    additional_run_info = {"error": "Timeout"}
                elif obj.exit_status is pynisher.MemorylimitException:
                    status = StatusType.MEMOUT
                    additional_run_info = {
                        "error": "Memout (used more than {} MB).".format(
                            self.memory_limit
                        )
                    }
                else:
                    raise ValueError(obj.exit_status)
                cost = self.worst_possible_result

        elif obj.exit_status is TAEAbortException:
            info = None
            status = StatusType.ABORT
            cost = self.worst_possible_result
            additional_run_info = {
                "error": "Your configuration of " "auto-sklearn does not work!",
                "exit_status": _encode_exit_status(obj.exit_status),
                "subprocess_stdout": obj.stdout,
                "subprocess_stderr": obj.stderr,
            }

        else:
            try:
                info = autosklearn.evaluation.util.read_queue(queue)
                result = info[-1]["loss"]
                status = info[-1]["status"]
                additional_run_info = info[-1]["additional_run_info"]

                if obj.exit_status == 0:
                    cost = result
                else:
                    status = StatusType.CRASHED
                    cost = self.worst_possible_result
                    additional_run_info["info"] = (
                        "Run treated as crashed "
                        "because the pynisher exit "
                        "status %s is unknown." % str(obj.exit_status)
                    )
                    additional_run_info["exit_status"] = _encode_exit_status(
                        obj.exit_status
                    )
                    additional_run_info["subprocess_stdout"] = obj.stdout
                    additional_run_info["subprocess_stderr"] = obj.stderr
            except Empty:
                info = None
                additional_run_info = {
                    "error": "Result queue is empty",
                    "exit_status": _encode_exit_status(obj.exit_status),
                    "subprocess_stdout": obj.stdout,
                    "subprocess_stderr": obj.stderr,
                    "exitcode": obj.exitcode,
                }
                status = StatusType.CRASHED
                cost = self.worst_possible_result

        if (
            self.budget_type is None or budget == 0
        ) and status == StatusType.DONOTADVANCE:
            status = StatusType.SUCCESS

        if not isinstance(additional_run_info, dict):
            additional_run_info = {"message": additional_run_info}

        if (
            info is not None
            and self.resampling_strategy
            in ("holdout-iterative-fit", "cv-iterative-fit")
            and status != StatusType.CRASHED
        ):
            learning_curve = autosklearn.evaluation.util.extract_learning_curve(info)
            learning_curve_runtime = autosklearn.evaluation.util.extract_learning_curve(
                info, "duration"
            )
            if len(learning_curve) > 1:
                additional_run_info["learning_curve"] = learning_curve
                additional_run_info["learning_curve_runtime"] = learning_curve_runtime

            train_learning_curve = autosklearn.evaluation.util.extract_learning_curve(
                info, "train_loss"
            )
            if len(train_learning_curve) > 1:
                additional_run_info["train_learning_curve"] = train_learning_curve
                additional_run_info["learning_curve_runtime"] = learning_curve_runtime

            if self._get_test_loss:
                test_learning_curve = (
                    autosklearn.evaluation.util.extract_learning_curve(
                        info,
                        "test_loss",
                    )
                )
                if len(test_learning_curve) > 1:
                    additional_run_info["test_learning_curve"] = test_learning_curve
                    additional_run_info[
                        "learning_curve_runtime"
                    ] = learning_curve_runtime

        if isinstance(config, int):
            origin = "DUMMY"
            config_id = config
        else:
            origin = getattr(config, "origin", "UNKNOWN")
            config_id = config.config_id

        additional_run_info["configuration_origin"] = origin

        runtime = float(obj.wall_clock_time)

        autosklearn.evaluation.util.empty_queue(queue)
        self.logger.info("Finished evaluating configuration %d" % config_id)

        # Do some sanity checking (for multi objective)
        if len(self.multi_objectives) > 1:
            error = (
                f"Returned costs {cost} does not match the number of objectives"
                f" {len(self.multi_objectives)}."
            )

            # If dict convert to array
            # Make sure the ordering is correct
            if isinstance(cost, dict):
                ordered_cost = []
                for name in self.multi_objectives:
                    if name not in cost:
                        raise RuntimeError(
                            f"Objective {name} was not found "
                            f"in the returned costs ({cost})"
                        )

                    ordered_cost.append(cost[name])
                cost = ordered_cost

            if isinstance(cost, list):
                if len(cost) != len(self.multi_objectives):
                    raise RuntimeError(error)

            if isinstance(cost, float):
                raise RuntimeError(error)

        return status, cost, runtime, additional_run_info



"""Test the output of loading the pareto set from an automl instance"""
from autosklearn.automl import AutoML

from pytest_cases import parametrize_with_cases
from pytest_cases.filters import has_tag

import test.test_automl.cases as cases

has_ensemble = has_tag("fitted") & ~has_tag("no_ensemble")

single_objective = has_ensemble & ~has_tag("multiobjective")
multi_objective = has_ensemble & has_tag("multiobjective")


@parametrize_with_cases("automl", cases=cases, filter=single_objective)
def test_can_output_pareto_front_singleobjective(automl: AutoML) -> None:
    """
    Expects
    -------
    * Non-multiobjective instances should have a pareto set of size 1
    """
    pareto_set = automl._load_pareto_set()

    assert len(pareto_set) == 1


@parametrize_with_cases("automl", cases=cases, filter=multi_objective)
def test_can_output_pareto_front_multiobjective(automl: AutoML) -> None:
    """
    Expects
    -------
    * Multiobjective ensembles should return >= 1, #TODO should test it's pareto optimal
    """
    pareto_set = automl._load_pareto_set()

    assert len(pareto_set) >= 1



from __future__ import annotations

from typing import Any, Callable, Dict, Tuple, Type

from functools import partial

from autosklearn.automl import AutoML, AutoMLClassifier, AutoMLRegressor
from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.ensembles.ensemble_selection import EnsembleSelection

from pytest import FixtureRequest, fixture
from unittest.mock import Mock

from test.conftest import DEFAULT_SEED
from test.fixtures.dask import create_test_dask_client


def _create_automl(
    automl_type: Type[AutoML] = AutoML,
    _id: str | None = None,
    **kwargs: Any,
) -> AutoML:
    """

    Parameters
    ----------
    automl_type : Type[AutoML] = AutoML
        The type of AutoML object to use

    _id: str | None = None
        If no dask client is provided, a unique id is required to create one
        so that it can be shut down after the test ends

    **kwargs: Any
        Options to pass on to the AutoML type for construction

    Returns
    -------
    AutoML
        The constructed class and a close method for dask, if it exists
    """
    test_defaults = {
        "n_jobs": 2,
        "time_left_for_this_task": 30,
        "per_run_time_limit": 5,
        "seed": DEFAULT_SEED,
        "ensemble_class": EnsembleSelection,
        "ensemble_kwargs": {"ensemble_size": 10},
        "ensemble_nbest": 10,
        "max_models_on_disc": 10,
        "initial_configurations_via_metalearning": 5,
    }

    # If a temp directory was explicitly passed, don't delete it automatically
    # Normally the `tmp_path` fixutre will delete it anyways
    if "temporary_directory" in kwargs:
        test_defaults["delete_tmp_folder_after_terminate"] = False

    opts: Dict[str, Any] = {**test_defaults, **kwargs}

    if "dask_client" not in opts:
        assert _id is not None
        client = create_test_dask_client(id=_id, n_workers=opts["n_jobs"])
        opts["dask_client"] = client

    auto = automl_type(**opts)
    return auto


@fixture
def make_automl(request: FixtureRequest) -> Callable[..., Tuple[AutoML, Callable]]:
    """See `_create_automl`"""
    yield partial(_create_automl, automl_type=AutoML, _id=request.node.nodeid)


@fixture
def make_automl_classifier(request: FixtureRequest) -> Callable[..., AutoMLClassifier]:
    """See `_create_automl`"""
    yield partial(_create_automl, automl_type=AutoMLClassifier, _id=request.node.nodeid)


@fixture
def make_automl_regressor(request: FixtureRequest) -> Callable[..., AutoMLRegressor]:
    """See `_create_automl`"""
    yield partial(_create_automl, automl_type=AutoMLRegressor, _id=request.node.nodeid)


class AutoMLStub(AutoML):
    def __init__(self) -> None:
        self.__class__ = AutoML
        self._task = None
        self._dask_client = None  # type: ignore
        self._is_dask_client_internally_created = False

    def __del__(self) -> None:
        pass


@fixture(scope="function")
def automl_stub() -> AutoMLStub:
    """TODO remove"""
    automl = AutoMLStub()
    automl._seed = 42
    automl._backend = Mock(spec=Backend)
    automl._backend.context = Mock()
    automl._delete_output_directories = lambda: 0
    return automl



from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling import (  # noqa: E501
    Rescaling,
)
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class QuantileTransformerComponent(Rescaling, AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        n_quantiles: int,
        output_distribution: str,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        from sklearn.preprocessing import QuantileTransformer

        self.n_quantiles = n_quantiles
        self.output_distribution = output_distribution
        self.preprocessor = QuantileTransformer(
            n_quantiles=n_quantiles,
            output_distribution=output_distribution,
            copy=False,
            random_state=random_state,
        )

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "QuantileTransformer",
            "name": "QuantileTransformer",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            # TODO find out if this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (INPUT, SIGNED_DATA),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        # TODO parametrize like the Random Forest as n_quantiles = n_features^param
        n_quantiles = UniformIntegerHyperparameter(
            "n_quantiles", lower=10, upper=2000, default_value=1000
        )
        output_distribution = CategoricalHyperparameter(
            "output_distribution", ["normal", "uniform"]
        )
        cs.add_hyperparameters((n_quantiles, output_distribution))
        return cs



import warnings

import numpy as np
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit
from sklearn.model_selection._split import _validate_shuffle_split
from sklearn.utils import _approximate_mode, check_random_state, indexable
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import _num_samples, check_array, column_or_1d


class CustomStratifiedShuffleSplit(StratifiedShuffleSplit):
    """Splitter that deals with classes with too few samples"""

    def _iter_indices(self, X, y, groups=None):  # type: ignore
        n_samples = _num_samples(X)
        y = check_array(y, ensure_2d=False, dtype=None)
        n_train, n_test = _validate_shuffle_split(
            n_samples,
            self.test_size,
            self.train_size,
            default_test_size=self._default_test_size,
        )

        if y.ndim == 2:
            # for multi-label y, map each distinct row to a string repr
            # using join because str(row) uses an ellipsis if len(row) > 1000
            y = np.array([" ".join(row.astype("str")) for row in y])

        classes, y_indices = np.unique(y, return_inverse=True)
        n_classes = classes.shape[0]

        class_counts = np.bincount(y_indices)

        if n_train < n_classes:
            raise ValueError(
                "The train_size = %d should be greater or "
                "equal to the number of classes = %d" % (n_train, n_classes)
            )
        if n_test < n_classes:
            raise ValueError(
                "The test_size = %d should be greater or "
                "equal to the number of classes = %d" % (n_test, n_classes)
            )

        # Find the sorted list of instances for each class:
        # (np.unique above performs a sort, so code is O(n logn) already)
        class_indices = np.split(
            np.argsort(y_indices, kind="mergesort"), np.cumsum(class_counts)[:-1]
        )

        rng = check_random_state(self.random_state)

        for _ in range(self.n_splits):
            # if there are ties in the class-counts, we want
            # to make sure to break them anew in each iteration
            n_i = _approximate_mode(class_counts, n_train, rng)
            class_counts_remaining = class_counts - n_i
            t_i = _approximate_mode(class_counts_remaining, n_test, rng)
            train = []
            test = []

            # NOTE: Adapting for unique instances
            #
            #   Each list n_i, t_i represent the list of class in the
            #   training_set and test_set resepectively.
            #
            #   n_i = [100, 100, 0, 3]  # 100 of class '0', 0 of class '2'
            #   t_i = [300, 300, 1, 3]  # 300 of class '0', 1 of class '2'
            #
            #  To support unique labels such as class '2', which only has one sample
            #  between both n_i and t_i, we need to make sure that n_i has at least
            #  one sample of all classes. There is also the extra check to ensure
            #  that the sizes stay the same.
            #
            #   n_i = [ 99, 100, 1, 3]  # 100 of class '0', 0 of class '2'
            #            |       ^
            #            v       |
            #   t_i = [301, 300, 0, 3]  # 300 of class '0', 1 of class '2'
            #
            for i, class_count in enumerate(n_i):
                if class_count == 0:
                    t_i[i] -= 1
                    n_i[i] += 1

                    j = np.argmax(n_i)
                    if n_i[j] == 1:
                        warnings.warn(
                            "Can't respect size requirements for split.",
                            " The training set must contain all of the unique"
                            " labels that exist in the dataset.",
                        )
                    else:
                        n_i[j] -= 1
                        t_i[j] += 1

            for i in range(n_classes):
                permutation = rng.permutation(class_counts[i])
                perm_indices_class_i = class_indices[i].take(permutation, mode="clip")

                train.extend(perm_indices_class_i[: n_i[i]])
                test.extend(perm_indices_class_i[n_i[i] : n_i[i] + t_i[i]])

            train = rng.permutation(train)
            test = rng.permutation(test)

            yield train, test


class CustomStratifiedKFold(StratifiedKFold):
    """Stratified K-Folds cross-validator that ensures that there is always at least
    1 sample per class in the training set.
    """

    def _make_test_folds(self, X, y=None):  # type: ignore
        rng = check_random_state(self.random_state)
        y = np.asarray(y)
        type_of_target_y = type_of_target(y)
        allowed_target_types = ("binary", "multiclass")
        if type_of_target_y not in allowed_target_types:
            raise ValueError(
                "Supported target types are: {}. Got {!r} instead.".format(
                    allowed_target_types, type_of_target_y
                )
            )

        y = column_or_1d(y)

        _, y_idx, y_inv = np.unique(y, return_index=True, return_inverse=True)
        # y_inv encodes y according to lexicographic order. We invert y_idx to
        # map the classes so that they are encoded by order of appearance:
        # 0 represents the first label appearing in y, 1 the second, etc.
        _, class_perm = np.unique(y_idx, return_inverse=True)
        y_encoded = class_perm[y_inv]

        n_classes = len(y_idx)

        # Determine the optimal number of samples from each class in each fold,
        # using round robin over the sorted y. (This can be done direct from
        # counts, but that code is unreadable.)
        y_order = np.sort(y_encoded)
        allocation = np.asarray(
            [
                np.bincount(y_order[i :: self.n_splits], minlength=n_classes)
                for i in range(self.n_splits)
            ]
        )

        # To maintain the data order dependencies as best as possible within
        # the stratification constraint, we assign samples from each class in
        # blocks (and then mess that up when shuffle=True).
        test_folds = np.empty(len(y), dtype="i")
        for k in range(n_classes):
            # since the kth column of allocation stores the number of samples
            # of class k in each test set, this generates blocks of fold
            # indices corresponding to the allocation for class k.
            folds_for_class = np.arange(self.n_splits).repeat(allocation[:, k])
            if self.shuffle:
                rng.shuffle(folds_for_class)
            test_folds[y_encoded == k] = folds_for_class
        return test_folds

    def split(self, X, y=None, groups=None):  # type: ignore

        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)
        if self.n_splits > n_samples:
            raise ValueError(
                f"Cannot have number of splits n_splits={self.n_splits} greater"
                f" than the number of samples: n_samples={n_samples}."
            )

        for train, test in super().split(X, y, groups):
            all_classes = np.unique(y)
            train_classes = np.unique(y[train])
            train = list(train)
            test = list(test)
            missing_classes = set(all_classes) - set(train_classes)
            if len(missing_classes) > 0:
                # print(missing_classes)
                for diff in missing_classes:
                    # print(len(train), len(test))
                    to_move = np.where(y[test] == diff)[0][0]
                    # print(y[test][to_move])
                    train = train + [test[to_move]]
                    del test[to_move]
                    # print(len(train), len(test))
            train = np.array(train, dtype=int)
            test = np.array(test, dtype=int)

            yield train, test



import sklearn.discriminant_analysis

from autosklearn.pipeline.components.classification.lda import LDA

from .test_base import BaseClassificationComponentTest


class LDAComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 1.0
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.5614481896257509
    res["default_iris_sparse"] = -1
    res["default_digits"] = 0.88585306618093507
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.9811778992106861
    res["default_digits_multilabel"] = 0.82204896441795205
    res["default_digits_multilabel_proba"] = 0.9833070018235553

    sk_mod = sklearn.discriminant_analysis.LinearDiscriminantAnalysis
    module = LDA



import dask.distributed
from distributed.utils_test import inc

from autosklearn.util.single_thread_client import SingleThreadedClient

import pytest


def test_single_thread_client_like_dask_client():
    single_thread_client = SingleThreadedClient()
    assert isinstance(single_thread_client, dask.distributed.Client)
    future = single_thread_client.submit(inc, 1)
    assert isinstance(future, dask.distributed.Future)
    assert future.done()
    assert future.result() == 2
    assert sum(single_thread_client.nthreads().values()) == 1
    single_thread_client.close()
    single_thread_client.shutdown()

    # Client/Futures are printed, so make sure str works
    # str calls __rpr__ which is the purpose of below check
    assert str(future) != ""
    assert str(single_thread_client) != ""

    with pytest.raises(NotImplementedError):
        single_thread_client.get_scheduler_logs()



import typing
from typing import Any

from pathlib import Path

import dask.distributed


class DummyFuture(dask.distributed.Future):
    """
    A class that mimics a distributed Future, the outcome of
    performing submit on a distributed client.
    """

    def __init__(self, result: typing.Any) -> None:
        self._result = result  # type: typing.Any

    def result(self, timeout: typing.Optional[int] = None) -> typing.Any:
        return self._result

    def cancel(self) -> None:
        pass

    def done(self) -> bool:
        return True

    def __repr__(self) -> str:
        return "DummyFuture: {}".format(self._result)

    def __del__(self) -> None:
        pass


class SingleThreadedClient(dask.distributed.Client):
    """
    A class to Mock the Distributed Client class, in case
    Auto-Sklearn is meant to run in the current Thread.
    """

    def __init__(self) -> None:

        # Raise a not implemented error if using a method from Client
        implemented_methods = [
            "submit",
            "close",
            "shutdown",
            "write_scheduler_file",
            "_get_scheduler_info",
            "nthreads",
        ]
        method_list = [
            func
            for func in dir(dask.distributed.Client)
            if callable(getattr(dask.distributed.Client, func))
            and not func.startswith("__")
        ]
        for method in method_list:
            if method in implemented_methods:
                continue
            setattr(self, method, self._unsupported_method)
        pass

    def _unsupported_method(self) -> None:
        raise NotImplementedError()

    def submit(
        self,
        func: typing.Callable,
        *args: typing.List,
        priority: int = 0,
        key: Any = None,
        workers: Any = None,
        resources: Any = None,
        retries: Any = None,
        fifo_timeout: Any = "100 ms",
        allow_other_workers: Any = False,
        actor: Any = False,
        actors: Any = False,
        pure: Any = None,
        **kwargs: Any,
    ) -> typing.Any:
        """
        Note
        ----
        The keyword arguments caught in `dask.distributed.Client` need to
        be specified here so they don't get passed in as ``**kwargs`` to the
        ``func``.
        """
        return DummyFuture(func(*args, **kwargs))

    def close(self) -> None:
        pass

    def shutdown(self) -> None:
        pass

    def write_scheduler_file(self, scheduler_file: str) -> None:
        Path(scheduler_file).touch()
        return

    def _get_scheduler_info(self) -> typing.Dict:
        return {
            "workers": ["127.0.0.1"],
            "type": "Scheduler",
        }

    def nthreads(self) -> typing.Dict:
        return {
            "127.0.0.1": 1,
        }

    def __repr__(self) -> str:
        return "SingleThreadedClient()"

    def __del__(self) -> None:
        pass
